#!/usr/bin/env python
########################################################################
# File :    dirac-sys-sendmail
# Author :  Matvey Sapunov
########################################################################

"""
Utility to send an e-mail using DIRAC notification service.

Arguments:
  Formated text message. The message consists of e-mail headers and e-mail body
  separated by two newline characters. Headers are key : value pairs separated
  by newline character. Meaningful headers are "To:", "From:", "Subject:".
  Other keys will be ommited.
  Message body is an arbitrary string.

Examples:
  $ dirac-sys-sendmail "From: source@email.com\\nTo: destination@email.com\\nSubject: Test\\n\\nMessage body"
  echo "From: source@email.com\\nSubject: Test\\n\\nMessage body" | dirac-sys-sendmail destination@email.com
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import socket
import sys
import os

from DIRAC import gLogger, exit as DIRACexit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient


@Script()
def main():
    _, args = Script.parseCommandLine(ignoreErrors=True)

    arg = "".join(args)

    if not len(arg) > 0:
        gLogger.error("Missing argument")
        DIRACexit(2)

    try:
        head, body = arg.split("\\n\\n")
    except Exception as x:
        head = "To: %s" % arg
        body = sys.stdin.read()

    try:
        tmp, body = body.split("\\n\\n")
        head = tmp + "\\n" + head
    except Exception as x:
        pass

    body = "".join(body.strip())

    try:
        headers = dict((i.strip(), j.strip()) for i, j in (item.split(":") for item in head.split("\\n")))
    except Exception:
        gLogger.error("Failed to convert string: %s to email headers" % head)
        DIRACexit(4)

    if "To" not in headers:
        gLogger.error("Failed to get 'To:' field from headers %s" % head)
        DIRACexit(5)
    to = headers["To"]

    origin = "%s@%s" % (os.getenv("LOGNAME", "dirac"), socket.getfqdn())
    if "From" in headers:
        origin = headers["From"]

    subject = "Sent from %s" % socket.getfqdn()
    if "Subject" in headers:
        subject = headers["Subject"]

    ntc = NotificationClient()
    print("sendMail(%s,%s,%s,%s,%s)" % (to, subject, body, origin, False))
    result = ntc.sendMail(to, subject, body, origin, localAttempt=False)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(6)

    DIRACexit(0)


if __name__ == "__main__":
    main()



""" ElasticDB is a base class used to connect an Elasticsearch database and manages queries.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.DIRACDB import DIRACDB
from DIRAC.Core.Utilities.ElasticSearchDB import ElasticSearchDB
from DIRAC.ConfigurationSystem.Client.Utilities import getElasticDBParameters


class ElasticDB(DIRACDB, ElasticSearchDB):
    """Class for interfacing DIRAC ES DB definitions to ES clusters"""

    ########################################################################
    def __init__(self, dbname, fullName, indexPrefix=""):
        """c'tor

        :param self: self reference
        :param str dbName: DIRAC name of the database for example: 'MonitoringDB'
        :param str fullName: The DIRAC full name of the database for example: 'Monitoring/MonitoringDB'
        :param str indexPrefix: it is the indexPrefix used to load all indexes
        """
        self.fullname = fullName

        result = getElasticDBParameters(fullName)
        if not result["OK"]:
            raise RuntimeError("Cannot get database parameters: %s" % result["Message"])

        dbParameters = result["Value"]
        self._dbHost = dbParameters["Host"]
        self._dbPort = dbParameters["Port"]
        # we can have db which does not have any authentication...
        self.__user = dbParameters.get("User", "")
        self.__dbPassword = dbParameters.get("Password", "")
        self.__useSSL = dbParameters.get("SSL", True)
        self.__useCRT = dbParameters.get("CRT", True)
        self.__ca_certs = dbParameters.get("ca_certs", None)
        self.__client_key = dbParameters.get("client_key", None)
        self.__client_cert = dbParameters.get("client_cert", None)

        super(ElasticDB, self).__init__(
            host=self._dbHost,
            port=self._dbPort,
            user=self.__user,
            password=self.__dbPassword,
            indexPrefix=indexPrefix,
            useSSL=self.__useSSL,
            useCRT=self.__useCRT,
            ca_certs=self.__ca_certs,
            client_key=self.__client_key,
            client_cert=self.__client_cert,
        )

        if not self._connected:
            raise RuntimeError("Can not connect to ES cluster %s, exiting..." % self.clusterName)

        self.log.info("================= ElasticSearch ==================")
        self.log.info("Host: %s " % self._dbHost)
        if self._dbPort:
            self.log.info("Port: %d " % self._dbPort)
        else:
            self.log.info("Port: Not specified, assuming URL points to right location")
        self.log.info(
            "Connecting with %s, %s:%s"
            % (
                "SSL" if self.__useSSL else "no SSL",
                self.__user if self.__user else "no user",
                "with password" if self.__dbPassword else "no password",
            )
        )
        self.log.info("ClusterName: %s   " % self.clusterName)
        self.log.info("==================================================")



#!/usr/bin/env python
########################################################################
# File :    dirac-proxy-info.py
# Author :  Adrian Casajus
########################################################################
"""
Print information about the current proxy.

Example:
  $ dirac-proxy-info
  subject      : /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar/CN=proxy/CN=proxy
  issuer       : /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar/CN=proxy
  identity     : /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar
  timeleft     : 23:53:55
  DIRAC group  : dirac_user
  path         : /tmp/x509up_u40885
  username     : vhamar
  VOMS         : True
  VOMS fqan    : ['/formation']
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import sys

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Utilities.ReturnValues import S_OK


class Params(object):

    proxyLoc = False
    vomsEnabled = True
    csEnabled = True
    steps = False
    checkValid = False
    checkClock = True
    uploadedInfo = False

    def showVersion(self, arg):
        print("Version:")
        print(" ", __RCSID__)
        sys.exit(0)
        return S_OK()

    def setProxyLocation(self, arg):
        self.proxyLoc = arg
        return S_OK()

    def disableVOMS(self, arg):
        self.vomsEnabled = False
        return S_OK()

    def disableCS(self, arg):
        self.csEnabled = False
        return S_OK()

    def showSteps(self, arg):
        self.steps = True
        return S_OK()

    def validityCheck(self, arg):
        self.checkValid = True
        return S_OK()

    def disableClockCheck(self, arg):
        self.checkClock = False
        return S_OK()

    def setManagerInfo(self, arg):
        self.uploadedInfo = True
        return S_OK()


@Script()
def main():
    params = Params()

    Script.registerSwitch("f:", "file=", "File to use as user key", params.setProxyLocation)
    Script.registerSwitch("i", "version", "Print version", params.showVersion)
    Script.registerSwitch("n", "novoms", "Disable VOMS", params.disableVOMS)
    Script.registerSwitch("v", "checkvalid", "Return error if the proxy is invalid", params.validityCheck)
    Script.registerSwitch("x", "nocs", "Disable CS", params.disableCS)
    Script.registerSwitch("e", "steps", "Show steps info", params.showSteps)
    Script.registerSwitch("j", "noclockcheck", "Disable checking if time is ok", params.disableClockCheck)
    Script.registerSwitch("m", "uploadedinfo", "Show uploaded proxies info", params.setManagerInfo)

    Script.disableCS()
    Script.parseCommandLine()

    from DIRAC.Core.Utilities.NTP import getClockDeviation
    from DIRAC import gLogger
    from DIRAC.Core.Security.ProxyInfo import getProxyInfo, getProxyStepsInfo
    from DIRAC.Core.Security.ProxyInfo import formatProxyInfoAsString, formatProxyStepsInfoAsString
    from DIRAC.Core.Security import VOMS
    from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
    from DIRAC.ConfigurationSystem.Client.Helpers import Registry

    if params.csEnabled:
        retVal = Script.enableCS()
        if not retVal["OK"]:
            print("Cannot contact CS to get user list")

    if params.checkClock:
        result = getClockDeviation()
        if result["OK"]:
            deviation = result["Value"]
            if deviation > 600:
                gLogger.error("Your host clock seems to be off by more than TEN MINUTES! Thats really bad.")
            elif deviation > 180:
                gLogger.error("Your host clock seems to be off by more than THREE minutes! Thats bad.")
            elif deviation > 60:
                gLogger.error("Your host clock seems to be off by more than a minute! Thats not good.")

    result = getProxyInfo(params.proxyLoc, not params.vomsEnabled)
    if not result["OK"]:
        gLogger.error(result["Message"])
        sys.exit(1)
    infoDict = result["Value"]
    gLogger.notice(formatProxyInfoAsString(infoDict))
    if not infoDict["isProxy"]:
        gLogger.error("==============================\n!!! The proxy is not valid !!!")

    if params.steps:
        gLogger.notice("== Steps extended info ==")
        chain = infoDict["chain"]
        stepInfo = getProxyStepsInfo(chain)["Value"]
        gLogger.notice(formatProxyStepsInfoAsString(stepInfo))

    def invalidProxy(msg):
        gLogger.error("Invalid proxy:", msg)
        sys.exit(1)

    if params.uploadedInfo:
        result = gProxyManager.getUserProxiesInfo()
        if not result["OK"]:
            gLogger.error("Could not retrieve the uploaded proxies info", result["Message"])
        else:
            uploadedInfo = result["Value"]
            if not uploadedInfo:
                gLogger.notice("== No proxies uploaded ==")
            if uploadedInfo:
                gLogger.notice("== Proxies uploaded ==")
                maxDNLen = 0
                maxGroupLen = 0
                for userDN in uploadedInfo:
                    maxDNLen = max(maxDNLen, len(userDN))
                    for group in uploadedInfo[userDN]:
                        maxGroupLen = max(maxGroupLen, len(group))
                gLogger.notice(" %s | %s | Until (GMT)" % ("DN".ljust(maxDNLen), "Group".ljust(maxGroupLen)))
                for userDN in uploadedInfo:
                    for group in uploadedInfo[userDN]:
                        gLogger.notice(
                            " %s | %s | %s"
                            % (
                                userDN.ljust(maxDNLen),
                                group.ljust(maxGroupLen),
                                uploadedInfo[userDN][group].strftime("%Y/%m/%d %H:%M"),
                            )
                        )

    if params.checkValid:
        if infoDict["secondsLeft"] == 0:
            invalidProxy("Proxy is expired")
        if params.csEnabled and not infoDict["validGroup"]:
            invalidProxy("Group %s is not valid" % infoDict["group"])
        if "hasVOMS" in infoDict and infoDict["hasVOMS"]:
            requiredVOMS = Registry.getVOMSAttributeForGroup(infoDict["group"])
            if "VOMS" not in infoDict or not infoDict["VOMS"]:
                invalidProxy("Unable to retrieve VOMS extension")
            if len(infoDict["VOMS"]) > 1:
                invalidProxy("More than one voms attribute found")
            if requiredVOMS not in infoDict["VOMS"]:
                invalidProxy(
                    "Unexpected VOMS extension %s. Extension expected for DIRAC group is %s"
                    % (infoDict["VOMS"][0], requiredVOMS)
                )
            result = VOMS.VOMS().getVOMSProxyInfo(infoDict["chain"], "actimeleft")
            if not result["OK"]:
                invalidProxy("Cannot determine life time of VOMS attributes: %s" % result["Message"])
            if int(result["Value"].strip()) == 0:
                invalidProxy("VOMS attributes are expired")

    sys.exit(0)


if __name__ == "__main__":
    main()



""" DIRAC FileCatalog mix-in class to manage users and groups within the FC database
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import time
import threading

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.UserGroupManager.UserAndGroupManagerBase import (
    UserAndGroupManagerBase,
)


class UserAndGroupManagerDB(UserAndGroupManagerBase):
    def getUserAndGroupID(self, credDict):
        """Get a uid, gid tuple for the given Credentials"""
        # Get the user
        s_uid = credDict.get("username", "anon")
        res = self.getUserID(s_uid)
        if not res["OK"]:
            return res
        uid = res["Value"]
        # Get the group (create it if it doesn't exist)
        s_gid = credDict.get("group", "anon")
        res = self.getGroupID(s_gid)
        if not res["OK"]:
            return res
        gid = res["Value"]
        return S_OK((uid, gid))

    #####################################################################
    #
    #  User related methods
    #
    #####################################################################

    def getUserID(self, user):
        """Get ID for a user specified by its name"""
        if isinstance(user, six.integer_types):
            return S_OK(user)
        if user in self.db.users.keys():
            return S_OK(self.db.users[user])
        return self.__addUser(user)

    def addUser(self, uname):
        """Add a new user with a name 'uname'"""
        return self.getUserID(uname)

    def getUsers(self):
        # self.__refreshUsers()
        return S_OK(self.db.users)

    def findUser(self, user):
        return self.getUserID(user)

    def getUserName(self, uid):
        """Get user name for the given id"""
        if uid in self.db.uids.keys():
            return S_OK(self.db.uids[uid])
        return S_ERROR("User id %d not found" % uid)

    def deleteUser(self, uname, force=True):
        """Delete a user specified by its name"""
        # ToDo: Check first if there are files belonging to the user
        if not force:
            pass
        return self.__removeUser(uname)

    def __addUser(self, uname):
        startTime = time.time()
        self.lock.acquire()
        waitTime = time.time()
        gLogger.debug("UserGroupManager AddUser lock created. Waited %.3f seconds. %s" % (waitTime - startTime, uname))
        if uname in self.db.users.keys():
            uid = self.db.users[uname]
            gLogger.debug(
                "UserGroupManager AddUser lock released. Used %.3f seconds. %s" % (time.time() - waitTime, uname)
            )
            self.lock.release()
            return S_OK(uid)
        res = self.db.insertFields("FC_Users", ["UserName"], [uname])
        if not res["OK"]:
            gLogger.debug(
                "UserGroupManager AddUser lock released. Used %.3f seconds. %s" % (time.time() - waitTime, uname)
            )
            self.lock.release()
            if "Duplicate entry" in res["Message"]:
                result = self._refreshUsers()
                if not result["OK"]:
                    return result
                if uname in self.db.users.keys():
                    uid = self.db.users[uname]
                    return S_OK(uid)
            return res
        uid = res["lastRowId"]
        self.db.uids[uid] = uname
        self.db.users[uname] = uid
        gLogger.debug("UserGroupManager AddUser lock released. Used %.3f seconds. %s" % (time.time() - waitTime, uname))
        self.lock.release()
        return S_OK(uid)

    def __removeUser(self, uname):
        startTime = time.time()
        self.lock.acquire()
        waitTime = time.time()
        gLogger.debug(
            "UserGroupManager RemoveUser lock created. Waited %.3f seconds. %s" % (waitTime - startTime, uname)
        )
        uid = self.db.users.get(uname, "Missing")
        req = "DELETE FROM FC_Users WHERE UserName='%s'" % uname
        res = self.db._update(req)
        if not res["OK"]:
            gLogger.debug(
                "UserGroupManager RemoveUser lock released. Used %.3f seconds. %s" % (time.time() - waitTime, uname)
            )
            self.lock.release()
            return res
        if uid != "Missing":
            self.db.users.pop(uname)
            self.db.uids.pop(uid)
        gLogger.debug(
            "UserGroupManager RemoveUser lock released. Used %.3f seconds. %s" % (time.time() - waitTime, uname)
        )
        self.lock.release()
        return S_OK()

    def _refreshUsers(self):
        """Get the current user IDs and names"""
        startTime = time.time()
        self.lock.acquire()
        waitTime = time.time()
        gLogger.debug("UserGroupManager RefreshUsers lock created. Waited %.3f seconds." % (waitTime - startTime))
        req = "SELECT UID,UserName from FC_Users"
        res = self.db._query(req)
        if not res["OK"]:
            gLogger.debug("UserGroupManager RefreshUsers lock released. Used %.3f seconds." % (time.time() - waitTime))
            self.lock.release()
            return res
        self.db.users = {}
        self.db.uids = {}
        for uid, uname in res["Value"]:
            self.db.users[uname] = uid
            self.db.uids[uid] = uname
        gLogger.debug("UserGroupManager RefreshUsers lock released. Used %.3f seconds." % (time.time() - waitTime))
        self.lock.release()
        return S_OK()

    #####################################################################
    #
    #  Group related methods
    #

    def getGroupID(self, group):
        """Get ID for a group specified by its name"""
        if isinstance(group, six.integer_types):
            return S_OK(group)
        if group in self.db.groups.keys():
            return S_OK(self.db.groups[group])
        return self.__addGroup(group)

    def addGroup(self, gname):
        """Add a new group with a name 'name'"""
        return self.getGroupID(gname)

    def getGroups(self):
        # self.__refreshGroups()
        return S_OK(self.db.groups)

    def findGroup(self, group):
        return self.getGroupID(group)

    def getGroupName(self, gid):
        """Get group name for the given id"""
        if gid in self.db.gids.keys():
            return S_OK(self.db.gids[gid])
        return S_ERROR("Group id %d not found" % gid)

    def deleteGroup(self, gname, force=True):
        """Delete a group specified by its name"""
        if not force:
            # ToDo: Check first if there are files belonging to the group
            pass
        return self.__removeGroup(gname)

    def __addGroup(self, group):
        startTime = time.time()
        self.lock.acquire()
        waitTime = time.time()
        gLogger.debug("UserGroupManager AddGroup lock created. Waited %.3f seconds. %s" % (waitTime - startTime, group))
        if group in self.db.groups.keys():
            gid = self.db.groups[group]
            gLogger.debug(
                "UserGroupManager AddGroup lock released. Used %.3f seconds. %s" % (time.time() - waitTime, group)
            )
            self.lock.release()
            return S_OK(gid)
        res = self.db.insertFields("FC_Groups", ["GroupName"], [group])
        if not res["OK"]:
            gLogger.debug(
                "UserGroupManager AddGroup lock released. Used %.3f seconds. %s" % (time.time() - waitTime, group)
            )
            self.lock.release()
            if "Duplicate entry" in res["Message"]:
                result = self._refreshGroups()
                if not result["OK"]:
                    return result
                if group in self.db.groups.keys():
                    gid = self.db.groups[group]
                    return S_OK(gid)
            return res
        gid = res["lastRowId"]
        self.db.gids[gid] = group
        self.db.groups[group] = gid
        gLogger.debug(
            "UserGroupManager AddGroup lock released. Used %.3f seconds. %s" % (time.time() - waitTime, group)
        )
        self.lock.release()
        return S_OK(gid)

    def __removeGroup(self, group):
        startTime = time.time()
        self.lock.acquire()
        waitTime = time.time()
        gLogger.debug(
            "UserGroupManager RemoveGroup lock created. Waited %.3f seconds. %s" % (waitTime - startTime, group)
        )
        gid = self.db.groups.get(group, "Missing")
        req = "DELETE FROM FC_Groups WHERE GroupName='%s'" % group
        res = self.db._update(req)
        if not res["OK"]:
            gLogger.debug(
                "UserGroupManager RemoveGroup lock released. Used %.3f seconds. %s" % (time.time() - waitTime, group)
            )
            self.lock.release()
            return res
        if gid != "Missing":
            self.db.groups.pop(group)
            self.db.gids.pop(gid)
        gLogger.debug(
            "UserGroupManager RemoveGroup lock released. Used %.3f seconds. %s" % (time.time() - waitTime, group)
        )
        self.lock.release()
        return S_OK()

    def _refreshGroups(self):
        """Get the current group IDs and names"""
        req = "SELECT GID,GroupName from FC_Groups"
        startTime = time.time()
        self.lock.acquire()
        waitTime = time.time()
        gLogger.debug("UserGroupManager RefreshGroups lock created. Waited %.3f seconds." % (waitTime - startTime))
        res = self.db._query(req)
        if not res["OK"]:
            gLogger.debug("UserGroupManager RefreshGroups lock released. Used %.3f seconds." % (time.time() - waitTime))
            self.lock.release()
            return res
        self.db.groups = {}
        self.db.gids = {}
        for gid, gname in res["Value"]:
            self.db.groups[gname] = gid
            self.db.gids[gid] = gname
        gLogger.debug("UserGroupManager RefreshGroups lock released. Used %.3f seconds." % (time.time() - waitTime))
        self.lock.release()
        return S_OK()



"""
    Workflow class is the main container of Steps and Modules
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import xml.sax
import six

from DIRAC.Core.Workflow.Parameter import *
from DIRAC.Core.Workflow.Module import *
from DIRAC.Core.Workflow.Step import *
from DIRAC.Core.Workflow.Utility import *
from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"


class Workflow(AttributeCollection):
    def __init__(self, obj=None, name=None):
        """Be aware that 1-st param is an obj not a name!!!!
        obj can me a string with XML representation or with filename
        also obj can be a Workflow or ParameterCollections
        """
        AttributeCollection.__init__(self)
        if (obj is None) or isinstance(obj, ParameterCollection):
            self.setName("notgiven")
            self.setType("")
            self.setDescrShort("")
            self.setDescription("")
            self.setOrigin("")
            self.setVersion(0.0)
            self.parameters = ParameterCollection(obj)
            self.step_instances = InstancesPool(self)
            self.step_definitions = DefinitionsPool(self)
            self.module_definitions = DefinitionsPool(self)

        elif isinstance(obj, Workflow):
            self.fromWorkflow(obj)
        elif isinstance(obj, six.string_types):
            self.parameters = ParameterCollection(None)
            self.step_instances = InstancesPool(self)
            self.step_definitions = DefinitionsPool(self)
            self.module_definitions = DefinitionsPool(self)
            # if obj is an XML string
            if obj.startswith("<"):
                fromXMLString(obj, self)
            else:
                fromXMLFile(obj, self)
        elif obj is not None:
            raise TypeError("Can not create object type " + str(type(self)) + " from the " + str(type(obj)))

        if name:
            self.setName(name)
        self.workflow_commons = {}
        self.workflowStatus = S_OK()

    def fromWorkflow(self, obj):
        self.setName(obj.getName())
        self.setType(obj.getType())
        self.setDescrShort(obj.getDescrShort())
        self.setDescription(obj.getDescription())
        self.setOrigin(obj.getOrigin())
        self.setVersion(obj.getVersion())
        # copy instances and definitions
        self.parameters = ParameterCollection(obj.parameters)
        self.module_definitions = DefinitionsPool(self, obj.module_definitions)
        self.step_instances = InstancesPool(self, obj.step_instances)
        self.step_definitions = DefinitionsPool(self, obj.step_definitions)

    def __str__(self):
        """Creates a string representation of itself"""
        ret = str(self.getName()) + ":\n" + AttributeCollection.__str__(self) + self.parameters.__str__()
        ret = ret + str(self.step_definitions)
        ret = ret + str(self.step_instances)
        ret = ret + str(self.module_definitions)
        return ret

    def toXML(self):
        """Creates an XML representation of itself"""
        # THIS is very important that Definitions should be written before instances
        ret = "<Workflow>\n"
        ret = ret + AttributeCollection.toXML(self)
        ret = ret + self.parameters.toXML()
        ret = ret + self.module_definitions.toXML()
        ret = ret + self.step_definitions.toXML()
        ret = ret + self.step_instances.toXML()
        ret = ret + "</Workflow>\n"
        return ret

    def toXMLFile(self, outFile):
        if os.path.exists(outFile):
            os.remove(outFile)
        with open(outFile, "w") as xmlfile:
            xmlfile.write(self.toXML())

    def addTool(self, name, tool):
        """Add an object that will be available in all the modules to perform some operations.
        For example, a state reporting facility.
        """
        self.workflow_commons[name] = tool

    def addStep(self, step):
        # this is a VERY important piece of code
        # we have to join all Modules definition from all added steps in the single dictionary
        # and we have to share this dictionary between all included steps
        # we also have to check versions of the modules and instances
        for type_o in step.module_definitions:
            if type_o not in self.module_definitions:
                self.module_definitions.append(step.module_definitions[type_o])
        self.step_definitions.append(step)
        del step.module_definitions  # we need to clean up all unwanted definitions
        step.module_definitions = None
        return step

    def addModule(self, module):
        # KGG We need to add code to update existing modules
        self.module_definitions.append(module)
        return module

    def createStepInstance(self, type_o, name):
        """Creates step instance of type 'type_o' with the name 'name'"""
        if type_o in self.step_definitions:
            stepi = StepInstance(name, self.step_definitions[type_o])
            self.step_instances.append(stepi)
            return stepi
        else:
            raise KeyError("Can not find StepDefinition " + type + " to create StepInstrance " + name)

    def removeStepInstance(self, name):
        self.step_instances[name].setParent(None)
        self.step_instances.delete(name)

    def updateParents(self):
        self.module_definitions.updateParents(self)
        self.step_instances.updateParents(self)
        self.step_definitions.updateParents(self)

    def resolveGlobalVars(self):
        """This method will create global parameter list and then will resolve all instances of @{VARNAME}
        Be aware that parameters of that type are GLOBAL!!! are string and can not be dynamically change
        The scope: the resolution of that parameters apply from lower to upper object, for example if
        parameter use in module, then it checks module, then step, then workflow

        Comment: If varible linked it should not be used in a global list"""

        # reenforced global parameters on the level of Workflow
        if not self.parameters.find("PRODUCTION_ID"):
            self.parameters.append(
                Parameter(
                    "PRODUCTION_ID",
                    "00000000",
                    "string",
                    "",
                    "",
                    True,
                    False,
                    "Transformation ID taken from the ProductionManager",
                )
            )
        if not self.parameters.find("JOB_ID"):
            self.parameters.append(
                Parameter(
                    "JOB_ID",
                    "00000000",
                    "string",
                    "",
                    "",
                    True,
                    False,
                    "Job ID within Tranformation taken from the Transformation Manager",
                )
            )

        self.parameters.resolveGlobalVars()
        step_instance_number = 0
        for inst in self.step_instances:
            # for each step instance we can define STEP_NUMBER
            step_instance_number = step_instance_number + 1
            if not inst.parameters.find("STEP_NUMBER"):
                inst.parameters.append(
                    Parameter(
                        "STEP_NUMBER",
                        "%s" % step_instance_number,
                        "string",
                        "",
                        "",
                        True,
                        False,
                        "Number of the StepInstance within the Workflow",
                    )
                )
            if not inst.parameters.find("STEP_ID"):
                prod_ID = self.parameters.find("PRODUCTION_ID").getValue()
                job_ID = self.parameters.find("JOB_ID").getValue()
                inst.parameters.append(
                    Parameter(
                        "STEP_ID",
                        "%s_%s_%d" % (prod_ID, job_ID, step_instance_number),
                        "string",
                        "",
                        "",
                        True,
                        False,
                        "Step instance ID",
                    )
                )
            if not inst.parameters.find("STEP_INSTANCE_NAME"):
                inst.parameters.append(
                    Parameter(
                        "STEP_INSTANCE_NAME",
                        inst.getName(),
                        "string",
                        "",
                        "",
                        True,
                        False,
                        "Name of the StepInstance within the Workflow",
                    )
                )
            if not inst.parameters.find("STEP_DEFINITION_NAME"):
                inst.parameters.append(
                    Parameter(
                        "STEP_DEFINITION_NAME",
                        inst.getType(),
                        "string",
                        "",
                        "",
                        True,
                        False,
                        "Type of the StepInstance within the Workflow",
                    )
                )
            if not inst.parameters.find("JOB_ID"):
                inst.parameters.append(
                    Parameter(
                        "JOB_ID",
                        "",
                        "string",
                        "self",
                        "JOB_ID",
                        True,
                        False,
                        "Job ID within Tranformation taken from the Transformation Manager",
                    )
                )
            if not inst.parameters.find("PRODUCTION_ID"):
                inst.parameters.append(
                    Parameter(
                        "PRODUCTION_ID",
                        "",
                        "string",
                        "self",
                        "PRODUCTION_ID",
                        True,
                        False,
                        "Type of the StepInstance within the Workflow",
                    )
                )
            inst.resolveGlobalVars(self.step_definitions, self.parameters)

    def createCode(self, combine_steps=False):
        self.resolveGlobalVars()
        str = ""
        str = str + self.module_definitions.createCode()
        str = str + self.step_definitions.createCode()
        str = str + "\nclass job:\n"
        str = str + indent(1) + "def execute(self):\n"
        # str=str+indent(2)+'# flush self.step_instances\n'
        str = str + self.step_instances.createCode()
        # it seems we do not need it on this level
        str = str + indent(2) + "# output assignment\n"
        for v in self.parameters:
            if v.isOutput():
                str = str + v.createParameterCode(2, "self")

        str = str + "\nj=job()\n"
        str = str + self.parameters.createParametersCode(0, "j")
        str = str + "j.execute()"
        return str

    def execute(self):
        self.resolveGlobalVars()
        # define workflow attributes
        wf_exec_attr = {}  # dictionary with the WF attributes, used to resolve links to self.attrname
        for wf_parameter in self.parameters:
            # parameters shall see objects in the current scope order to resolve links
            if wf_parameter.preExecute():  # for parm which not just outputs
                # print 'Input', wf_parameter
                if wf_parameter.isLinked():
                    # print "Workflow self." + wf_parameter.getName(), '=',
                    # wf_parameter.getLinkedModule() + '.' + wf_parameter.getLinkedParameter()
                    if wf_parameter.getLinkedModule() == "self":
                        # this is not suppose to happen
                        print(
                            "Warning! Job attribute ",
                            wf_parameter.getName(),
                            "refers to the attribute of the same workflow",
                            wf_parameter.getLinkedParameter(),
                        )
                        wf_exec_attr[wf_parameter.getName()] = wf_exec_attr[wf_parameter.getLinkedParameter()]
                    else:
                        wf_exec_attr[wf_parameter.getName()] = wf_exec_attr[wf_parameter.getLinkedModule()][
                            wf_parameter.getLinkedParameter()
                        ]
                else:
                    # print "Workflow self." + wf_parameter.getName(), '=', wf_parameter.getValue()
                    wf_exec_attr[wf_parameter.getName()] = wf_parameter.getValue()

            # Put all the workflow parameters into the workflow_commons dictionary
            self.workflow_commons[wf_parameter.getName()] = wf_parameter.getValue()

        self.module_definitions.loadCode()  # loading Module classes into current python scope

        # wf_exec_steps will be dictionary of dictionaries [step instance name][parameter name]
        # used as dictionary of step instances to carry parameters
        wf_exec_steps = {}
        # print 'Executing Workflow',self.getType()
        error_message = ""
        step_result = ""
        for step_inst in self.step_instances:
            step_inst_name = step_inst.getName()
            step_inst_type = step_inst.getType()
            wf_exec_steps[step_inst_name] = {}
            # print "WorkflowInstance creating Step instance ",step_inst_name," of type", step_inst_type
            for parameter in step_inst.parameters:
                if parameter.preExecute():
                    # print '>> Input', parameter
                    if parameter.isLinked():
                        # print ">> StepInstance", step_inst_name + '.' + parameter.getName(),
                        # '=', parameter.getLinkedModule() + '.' + parameter.getLinkedParameter()
                        if parameter.getLinkedModule() == "self":
                            # tale value form the step_dict
                            wf_exec_steps[step_inst_name][parameter.getName()] = wf_exec_attr[
                                parameter.getLinkedParameter()
                            ]
                        else:
                            # print wf_exec_steps[parameter.getLinkedModule()].keys()
                            wf_exec_steps[step_inst_name][parameter.getName()] = wf_exec_steps[
                                parameter.getLinkedModule()
                            ][parameter.getLinkedParameter()]
                    else:
                        # print ">> StepInstance", step_inst_name + '.' + parameter.getName(), '=', parameter.getValue()
                        wf_exec_steps[step_inst_name][parameter.getName()] = parameter.getValue()

                # In the step_commons all parameters are added, both Input and Output ones.
                step_inst.step_commons[parameter.getName()] = parameter.getValue()

            resolveVariables(wf_exec_steps[step_inst_name])
            # Set proper values for all Input Parameters
            for key, value in wf_exec_steps[step_inst_name].items():
                step_inst.step_commons[key] = value

            step_inst.setParent(self)
            step_inst.setWorkflowCommons(self.workflow_commons)

            result = step_inst.execute(wf_exec_steps[step_inst_name], self.step_definitions)
            if not result["OK"]:
                if self.workflowStatus["OK"]:
                    error_message = result["Message"]
                self.workflowStatus = S_ERROR(result["Message"])
                self.workflowStatus["Errno"] = result["Errno"]

            step_result = result.get("Value", step_result)

        # now we need to copy output values to the STEP!!! parameters
        # print "WorkflowInstance output assignment"
        for wf_parameter in self.parameters:
            if wf_parameter.isOutput():
                if wf_parameter.isLinked():
                    # print "WorkflowInstance  self." + wf_parameter.getName(), '=',
                    # wf_parameter.getLinkedModule() + '.' + wf_parameter.getLinkedParameter()
                    if wf_parameter.getLinkedModule() == "self":
                        # this is not suppose to happen
                        print(
                            "Warning! Workflow OUTPUT attribute ",
                            wf_parameter.getName(),
                            "refer on the attribute of the same workflow",
                            wf_parameter.getLinkedParameter(),
                        )
                        wf_exec_attr[wf_parameter.getName()] = wf_exec_attr[wf_parameter.getLinkedParameter()]
                    else:
                        wf_exec_attr[wf_parameter.getName()] = wf_exec_steps[wf_parameter.getLinkedModule()][
                            wf_parameter.getLinkedParameter()
                        ]
                else:
                    # it is also does not make sense - we can produce warning
                    print(
                        "Warning! Workflow OUTPUT attribute",
                        wf_parameter.getName(),
                        "assigned constant",
                        wf_parameter.getValue(),
                    )
                    # print "WorkflowInstance  self."+ wf_parameter.getName(),'=',wf_parameter.getValue()
                    wf_exec_attr[wf_parameter.getName()] = wf_parameter.getValue()
                setattr(self, wf_parameter.getName(), wf_exec_attr[wf_parameter.getName()])
        # Return the result of the first failed step or S_OK
        if not self.workflowStatus["OK"]:
            # return S_ERROR( error_message )
            return self.workflowStatus
        return S_OK(step_result)


from DIRAC.Core.Workflow.WorkflowReader import WorkflowXMLHandler


def fromXMLString(xml_string, obj=None):
    # KGG !!! We need to reset Workflow if it exists
    handler = WorkflowXMLHandler(obj)
    xml.sax.parseString(xml_string, handler)
    return handler.root


def fromXMLFile(xml_file, obj=None):
    # KGG !!! We need to reset Workflow if it exists
    handler = WorkflowXMLHandler(obj)
    xml.sax.parse(xml_file, handler)
    return handler.root



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import time
import copy

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.AccountingSystem.private.DBUtils import DBUtils
from DIRAC.Core.Utilities.Plotting import gDataCache
from DIRAC.Core.Utilities.Plotting.Plots import (
    generateNoDataPlot,
    generateTimedStackedBarPlot,
    generateQualityPlot,
    generateCumulativePlot,
    generatePiePlot,
    generateStackedLinePlot,
    generateHistogram,
)


class BaseReporter(DBUtils):

    _PARAM_CHECK_FOR_NONE = "checkNone"
    _PARAM_CALCULATE_PROPORTIONAL_GAUGES = "calculateProportionalGauges"
    _PARAM_CONVERT_TO_GRANULARITY = "convertToGranularity"
    _VALID_PARAM_CONVERT_TO_GRANULARITY = ("sum", "average")
    _PARAM_CONSOLIDATION_FUNCTION = "consolidationFunction"

    _EA_THUMBNAIL = "thumbnail"
    _EA_WIDTH = "width"
    _EA_HEIGHT = "height"
    _EA_THB_WIDTH = "thbWidth"
    _EA_THB_HEIGHT = "thbHeight"
    _EA_PADDING = "figurePadding"
    _EA_TITLE = "plotTitle"

    _RATE_UNITS = {
        "time": (
            ("seconds / s", 1, 24),
            ("hours / s", 3600, 24),
            ("days / s", 86400, 15),
            ("weeks / s", 86400 * 7, 10),
            ("months / s", 86400 * 30, 12),
            ("years / s", 86400 * 365, 1),
        ),
        "cpupower": (("HS06", 1, 750), ("kHS06", 1000, 750), ("MHS06", 10 ** 6, 1)),
        "bytes": (
            ("MB / s", 10 ** 6, 1024),
            ("GB / s", 10 ** 9, 1024),
            ("TB / s", 10 ** 12, 1024),
            ("PB / s", 10 ** 15, 1),
        ),
        "jobs": (
            ("jobs / hour", 1 / 3600.0, 1000),
            ("kjobs / hour", (10 ** 3) / 3600.0, 1000),
            ("Mjobs / hour", (10 ** 6) / 3600.0, 1),
        ),
        "files": (
            ("files / hour", 1 / 3600.0, 1000),
            ("kfiles / hour", (10 ** 3) / 3600.0, 1000),
            ("Mfiles / hour", (10 ** 6) / 3600.0, 1),
        ),
    }

    _UNITS = {
        "time": (
            ("seconds", 1, 24),
            ("hours", 3600, 24),
            ("days", 86400, 15),
            ("weeks", 86400 * 7, 10),
            ("months", 86400 * 30, 12),
            ("years", 86400 * 365, 1),
        ),
        "cpupower": (
            ("HS06 hours", 3600, 24),
            ("HS06 days", 86400, 750),
            ("kHS06 days", 86400 * 1000, 750),
            ("MHS06 days", 86400 * 10 ** 6, 1),
        ),
        "bytes": (("MB", 10 ** 6, 1024), ("GB", 10 ** 9, 1024), ("TB", 10 ** 12, 1024), ("PB", 10 ** 15, 1)),
        "jobs": (("jobs", 1, 1000), ("kjobs", 10 ** 3, 1000), ("Mjobs", 10 ** 6, 1)),
        "files": (("files", 1, 1000), ("kfiles", 10 ** 3, 1000), ("Mfiles", 10 ** 6, 1)),
    }

    # To be defined in the derived classes
    _typeKeyFields = []
    _typeName = ""

    def __init__(self, db, setup, extraArgs=None):
        DBUtils.__init__(self, db, setup)
        if isinstance(extraArgs, dict):
            self._extraArgs = extraArgs
        else:
            self._extraArgs = {}
        reportsRevMap = {}
        for attr in dir(self):
            if attr.find("_report") == 0:
                if attr.find("Name", len(attr) - 4) == len(attr) - 4:
                    reportId = attr[7:-4]
                    reportName = getattr(self, attr)
                    reportsRevMap[reportId] = reportName
                else:
                    reportId = attr[7:]
                    if reportId not in reportsRevMap:
                        reportsRevMap[reportId] = reportId
        self.__reportNameMapping = {}
        for rId in reportsRevMap:
            self.__reportNameMapping[reportsRevMap[rId]] = rId

    def _translateGrouping(self, grouping):
        return ("%s", [grouping])

    def _averageConsolidation(self, total, count):
        if count == 0:
            return 0
        else:
            return float(total) / float(count)

    def _efficiencyConsolidation(self, total, count):
        if count == 0:
            return 0
        else:
            return (float(total) / float(count)) * 100.0

    def generate(self, reportRequest):
        reportRequest["groupingFields"] = self._translateGrouping(reportRequest["grouping"])
        reportHash = reportRequest["hash"]
        reportName = reportRequest["reportName"]
        if reportName in self.__reportNameMapping:
            reportRequest["reportName"] = self.__reportNameMapping[reportName]
        gLogger.info("Retrieving data for %s:%s" % (reportRequest["typeName"], reportRequest["reportName"]))
        sT = time.time()
        retVal = self.__retrieveReportData(reportRequest, reportHash)
        reportGenerationTime = time.time() - sT
        if not retVal["OK"]:
            return retVal
        if not reportRequest["generatePlot"]:
            return retVal
        reportData = retVal["Value"]
        gLogger.info("Plotting data for %s:%s" % (reportRequest["typeName"], reportRequest["reportName"]))
        sT = time.time()
        retVal = self.__generatePlotForReport(reportRequest, reportHash, reportData)
        plotGenerationTime = time.time() - sT
        gLogger.info(
            "Time for %s:%s - Report %.2f Plot %.2f (%.2f%% r/p)"
            % (
                reportRequest["typeName"],
                reportRequest["reportName"],
                reportGenerationTime,
                plotGenerationTime,
                ((reportGenerationTime * 100 / plotGenerationTime) if plotGenerationTime else 0.0),
            )
        )
        if not retVal["OK"]:
            return retVal
        plotDict = retVal["Value"]
        if "retrieveReportData" in reportRequest["extraArgs"] and reportRequest["extraArgs"]["retrieveReportData"]:
            plotDict["reportData"] = reportData
        return S_OK(plotDict)

    def plotsList(self):
        return sorted([k for k in self.__reportNameMapping])

    def __retrieveReportData(self, reportRequest, reportHash):
        funcName = "_report%s" % reportRequest["reportName"]
        try:
            funcObj = getattr(self, funcName)
        except Exception:
            return S_ERROR("Report %s is not defined" % reportRequest["reportName"])
        return gDataCache.getReportData(reportRequest, reportHash, funcObj)

    def __generatePlotForReport(self, reportRequest, reportHash, reportData):
        funcName = "_plot%s" % reportRequest["reportName"]
        try:
            funcObj = getattr(self, funcName)
        except Exception:
            return S_ERROR("Plot function for report %s is not defined" % reportRequest["reportName"])
        return gDataCache.getReportPlot(reportRequest, reportHash, reportData, funcObj)

    ###
    # Helper functions for reporters
    ###

    def _getTimedData(self, startTime, endTime, selectFields, preCondDict, groupingFields, metadataDict):
        condDict = {}
        # Check params
        if self._PARAM_CHECK_FOR_NONE not in metadataDict:
            metadataDict[self._PARAM_CHECK_FOR_NONE] = False
        if self._PARAM_CONVERT_TO_GRANULARITY not in metadataDict:
            metadataDict[self._PARAM_CONVERT_TO_GRANULARITY] = "sum"
        elif metadataDict[self._PARAM_CONVERT_TO_GRANULARITY] not in self._VALID_PARAM_CONVERT_TO_GRANULARITY:
            return S_ERROR("%s field metadata is invalid" % self._PARAM_CONVERT_TO_GRANULARITY)
        if self._PARAM_CALCULATE_PROPORTIONAL_GAUGES not in metadataDict:
            metadataDict[self._PARAM_CALCULATE_PROPORTIONAL_GAUGES] = False
        # Make safe selections
        for keyword in self._typeKeyFields:
            if keyword in preCondDict:
                condDict[keyword] = preCondDict[keyword]
        # Query!
        timeGrouping = ("%%s, %s" % groupingFields[0], ["startTime"] + groupingFields[1])
        retVal = self._retrieveBucketedData(
            self._typeName, startTime, endTime, selectFields, condDict, timeGrouping, ("%s", ["startTime"])
        )
        if not retVal["OK"]:
            return retVal
        dataDict = self._groupByField(0, retVal["Value"])
        coarsestGranularity = self._getBucketLengthForTime(self._typeName, startTime)
        # Transform!
        for keyField in dataDict:
            if metadataDict[self._PARAM_CHECK_FOR_NONE]:
                dataDict[keyField] = self._convertNoneToZero(dataDict[keyField])
            if metadataDict[self._PARAM_CONVERT_TO_GRANULARITY] == "average":
                dataDict[keyField] = self._averageToGranularity(coarsestGranularity, dataDict[keyField])
            if metadataDict[self._PARAM_CONVERT_TO_GRANULARITY] == "sum":
                dataDict[keyField] = self._sumToGranularity(coarsestGranularity, dataDict[keyField])
            if self._PARAM_CONSOLIDATION_FUNCTION in metadataDict:
                dataDict[keyField] = self._executeConsolidation(
                    metadataDict[self._PARAM_CONSOLIDATION_FUNCTION], dataDict[keyField]
                )
        if metadataDict[self._PARAM_CALCULATE_PROPORTIONAL_GAUGES]:
            dataDict = self._calculateProportionalGauges(dataDict)
        return S_OK((dataDict, coarsestGranularity))

    def _executeConsolidation(self, functor, dataDict):
        for timeKey in dataDict:
            dataDict[timeKey] = [functor(*dataDict[timeKey])]
        return dataDict

    def _getSummaryData(
        self, startTime, endTime, selectFields, preCondDict, groupingFields, metadataDict=None, reduceFunc=False
    ):
        condDict = {}
        # Make safe selections
        for keyword in self._typeKeyFields:
            if keyword in preCondDict:
                condDict[keyword] = preCondDict[keyword]
        # Query!
        retVal = self._retrieveBucketedData(
            self._typeName, startTime, endTime, selectFields, condDict, groupingFields, []
        )
        if not retVal["OK"]:
            return retVal
        dataDict = self._groupByField(0, retVal["Value"])
        for key in dataDict:
            if not reduceFunc:
                dataDict[key] = dataDict[key][0][0]
            else:
                dataDict[key] = reduceFunc(*dataDict[key][0])
        # HACK to allow serialization of the type because MySQL decides to return data in a non python standard format
        for key in dataDict:
            dataDict[key] = float(dataDict[key])
        return S_OK(dataDict)

    def _getBucketData(self, startTime, endTime, selectFields, preCondDict):
        """
        It retrieves data for histogram.

        :param int startTime: epoch time
        :param int endTime: epoch time
        :param list selectFields: the value being plotted
        :param dict preCondDict: plot attributes (conditions)

        """

        condDict = {}
        # Make safe selections
        for keyword in self._typeKeyFields:
            if keyword in preCondDict:
                condDict[keyword] = preCondDict[keyword]
        retVal = self._retrieveBucketedData(self._typeName, startTime, endTime, selectFields, condDict)
        if not retVal["OK"]:
            return retVal
        dataDict = self._groupByField(0, retVal["Value"])
        data = {}
        # convert values to the correct format
        for key, values in dataDict.items():
            data[key] = [float(i[0]) for i in values]
        return S_OK(data)

    def _getSelectStringForGrouping(self, groupingFields):
        if len(groupingFields) == 3:
            return groupingFields[2]
        if len(groupingFields[1]) == 1:
            # If there's only one field, then we send the sql representation in pos 0
            return groupingFields[0]
        else:
            return "CONCAT( %s )" % ", ".join(["%s, '-'" % sqlRep for sqlRep in groupingFields[0]])

    def _findSuitableRateUnit(self, dataDict, maxValue, unit):
        return self._findUnitMagic(dataDict, maxValue, unit, self._RATE_UNITS)

    def _findSuitableUnit(self, dataDict, maxValue, unit):
        return self._findUnitMagic(dataDict, maxValue, unit, self._UNITS)

    def _findUnitMagic(self, reportDataDict, maxValue, unit, selectedUnits):
        if unit not in selectedUnits:
            raise AttributeError("%s is not a known rate unit" % unit)
        baseUnitData = selectedUnits[unit][0]
        if "staticUnits" in self._extraArgs and self._extraArgs["staticUnits"]:
            unitData = selectedUnits[unit][0]
        else:
            unitList = selectedUnits[unit]
            unitIndex = -1
            for unitName, unitDivFactor, unitThreshold in unitList:
                unitIndex += 1
                if maxValue / unitDivFactor < unitThreshold:
                    break
            unitData = selectedUnits[unit][unitIndex]
        # Apply divFactor to all units
        graphDataDict, maxValue = self._divideByFactor(copy.deepcopy(reportDataDict), unitData[1])
        if unitData == baseUnitData:
            reportDataDict = graphDataDict
        else:
            reportDataDict, dummyMaxValue = self._divideByFactor(reportDataDict, baseUnitData[1])
        return reportDataDict, graphDataDict, maxValue, unitData[0]

    ##
    # Plotting
    ##

    def __checkPlotMetadata(self, metadata):
        if self._EA_WIDTH in self._extraArgs and self._extraArgs[self._EA_WIDTH]:
            try:
                metadata[self._EA_WIDTH] = min(1600, max(200, int(self._extraArgs[self._EA_WIDTH])))
            except Exception:
                pass
        if self._EA_HEIGHT in self._extraArgs and self._extraArgs[self._EA_HEIGHT]:
            try:
                metadata[self._EA_HEIGHT] = min(1600, max(200, int(self._extraArgs[self._EA_HEIGHT])))
            except Exception:
                pass
        if self._EA_TITLE in self._extraArgs and self._extraArgs[self._EA_TITLE]:
            metadata["title"] = self._extraArgs[self._EA_TITLE]

    def __checkThumbnailMetadata(self, metadata):
        if self._EA_THUMBNAIL in self._extraArgs and self._extraArgs[self._EA_THUMBNAIL]:
            thbMD = dict(metadata)
            thbMD["legend"] = False
            if self._EA_THB_HEIGHT in self._extraArgs:
                thbMD[self._EA_HEIGHT] = self._extraArgs[self._EA_THB_HEIGHT]
            else:
                thbMD[self._EA_HEIGHT] = 125
            if self._EA_THB_WIDTH in self._extraArgs:
                thbMD[self._EA_WIDTH] = self._extraArgs[self._EA_THB_WIDTH]
            else:
                thbMD[self._EA_WIDTH] = 200
            thbMD[self._EA_PADDING] = 20
            for key in ("title", "ylabel", "xlabel"):
                if key in thbMD:
                    del thbMD[key]
            return thbMD
        return False

    def __plotData(self, filename, dataDict, metadata, funcToPlot):
        self.__checkPlotMetadata(metadata)
        if not dataDict:
            funcToPlot = generateNoDataPlot
        plotFileName = "%s.png" % filename
        finalResult = funcToPlot(plotFileName, dataDict, metadata)
        if not finalResult["OK"]:
            return finalResult
        thbMD = self.__checkThumbnailMetadata(metadata)
        if not thbMD:
            return S_OK({"plot": True, "thumbnail": False})
        thbFilename = "%s.thb.png" % filename
        retVal = funcToPlot(thbFilename, dataDict, thbMD)
        if not retVal["OK"]:
            return retVal
        return S_OK({"plot": True, "thumbnail": True})

    def _generateTimedStackedBarPlot(self, filename, dataDict, metadata):
        return self.__plotData(filename, dataDict, metadata, generateTimedStackedBarPlot)

    def _generateQualityPlot(self, filename, dataDict, metadata):
        return self.__plotData(filename, dataDict, metadata, generateQualityPlot)

    def _generateCumulativePlot(self, filename, dataDict, metadata):
        return self.__plotData(filename, dataDict, metadata, generateCumulativePlot)

    def _generatePiePlot(self, filename, dataDict, metadata):
        return self.__plotData(filename, dataDict, metadata, generatePiePlot)

    def _generateStackedLinePlot(self, filename, dataDict, metadata):
        return self.__plotData(filename, dataDict, metadata, generateStackedLinePlot)

    def _generateHistogram(self, filename, dataDict, metadata):
        return self.__plotData(filename, dataDict, metadata, generateHistogram)



"""
This class is used to insert data to a db (currently elasticsearch).
It uses an internal list which is used to keep messages in the memory.
addRecord is used to insert messages to the internal queue. commit is used
to insert the acumulated messages to elasticsearch.
It provides two failover mechanism:
1.) If the database is not available, the data will be kept in the memory.
2.) If a MQ is available, we store the messages in MQ service.

Note: In order to not send too many rows to the db we use  __maxRecordsInABundle.

**Configuration Parameters**:

If the section ``/Resources/MQServices`` exists in the CS, it's assumed that
a MQ service is available, if the MQ is not working a failover will be performed.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import threading
import json

from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.Resources.MessageQueue.MQCommunication import createConsumer
from DIRAC.Resources.MessageQueue.MQCommunication import createProducer
from DIRAC.MonitoringSystem.Client.ServerUtils import monitoringDB
from DIRAC.ConfigurationSystem.Client.Config import gConfig


class MonitoringReporter(object):

    """
    .. class:: MonitoringReporter

    This class is used to interact with the ES DB, using a MQ as a failover mechanism.

    :param int __maxRecordsInABundle: limit the number of records to be inserted to the db.
    :param threading.RLock __documentLock: is used to lock the local store when it is being modified.
    :param list __documents: contains the records which will be inserted to the db.
    :param str __monitoringType: type of the records which will be inserted to the db. For example: WMSHistory.
    :param str __failoverQueueName: the name of the messaging queue. For example: /queue/dirac.certification
    """

    def __init__(self, monitoringType="", failoverQueueName="dirac.monitoring"):

        self.__maxRecordsInABundle = 5000
        self.__documentLock = threading.RLock()
        self.__documents = []
        self.__monitoringType = monitoringType
        self.__failoverQueueName = failoverQueueName
        self.__defaultMQProducer = None

    def __del__(self):
        if self.__defaultMQProducer is not None:
            self.__defaultMQProducer.close()

    def processRecords(self):
        """
        It consumes all messages from the MQ (these are failover messages).
        In case of failure, the messages will be inserted to the MQ again.
        """
        retVal = monitoringDB.pingDB()  # if the db is not accessible, the records will be not processed from MQ
        if retVal["OK"]:
            if not retVal["Value"]:  # false if we can not connect to the db
                return retVal
        else:
            return retVal

        result = createConsumer("Monitoring::Queues::%s" % self.__failoverQueueName)
        if not result["OK"]:
            gLogger.error("Fail to create Consumer", result["Message"])
            return S_ERROR("Fail to create Consumer")
        mqConsumer = result["Value"]

        result = S_OK()
        failedToProcess = []
        while result["OK"]:
            # we consume all messages from the consumer internal queue.
            result = mqConsumer.get()
            if result["OK"]:
                records = json.loads(result["Value"])
                retVal = monitoringDB.put(list(records), self.__monitoringType)
                if not retVal["OK"]:
                    failedToProcess.append(records)

        mqConsumer.close()  # make sure that we will not process any more messages.
        # the db is not available and we publish again the data to MQ
        for records in failedToProcess:
            res = self.publishRecords(records)
            if not res["OK"]:
                return res
        return S_OK()

    def addRecord(self, rec):
        """
        It inserts the record to the list

        :param dict rec: it contains a key/value pair.
        """
        self.__documentLock.acquire()
        self.__documents.append(rec)
        self.__documentLock.release()

    def publishRecords(self, records, mqProducer=None):
        """
        send data to the MQ. If the mqProducer instance is provided, it will be used for publishing the data to MQ.

        :param list records: contains a list of key/value pairs (dictionaries)
        :param object mqProducer: We can provide the instance of a producer, which will be used to publish the data
        """

        # If no mqProducer provided, we try to get the default one to send those records.
        if mqProducer is None:
            mqProducer = self.__getProducer()
            if mqProducer is None:
                gLogger.error("Fail to get Producer")
                return S_ERROR("Fail to get Producer")
            return mqProducer.put(json.dumps(records))

        return mqProducer.put(json.dumps(records))

    def commit(self):
        """
        It inserts the accumulated data to the db.
        In case of failure it keeps in memory/MQ
        """
        # before we try to insert the data to the db, we process all the data
        # which are already in the queue
        mqProducer = self.__getProducer()  # we are sure that we can connect to MQ
        if mqProducer is not None:
            result = self.processRecords()
            if not result["OK"]:
                gLogger.error("Unable to insert data to the db:", result["Message"])

        self.__documentLock.acquire()
        documents = self.__documents
        self.__documents = []
        self.__documentLock.release()
        recordSent = 0
        try:
            while documents:
                recordsToSend = documents[: self.__maxRecordsInABundle]
                retVal = monitoringDB.put(recordsToSend, self.__monitoringType)
                if retVal["OK"]:
                    recordSent += len(recordsToSend)
                    del documents[: self.__maxRecordsInABundle]
                    gLogger.verbose("%d records inserted to MonitoringDB" % (recordSent))
                else:
                    if mqProducer is not None:
                        res = self.publishRecords(recordsToSend, mqProducer)
                        if not res["OK"]:  # in case of MQ problem
                            return res
                        # if we managed to publish the records we can delete from the list
                        recordSent += len(recordsToSend)
                        del documents[: self.__maxRecordsInABundle]
                    else:
                        gLogger.warn("Failed to insert the records:", retVal["Message"])
        except Exception as e:  # pylint: disable=broad-except
            gLogger.exception("Error committing", lException=e)
            return S_ERROR("Error committing %s" % repr(e).replace(",)", ")"))
        finally:
            self.__documentLock.acquire()
            self.__documents.extend(documents)
            self.__documentLock.release()
        return S_OK(recordSent)

    def __getProducer(self):
        """
        This method is used to get the default MQ producer or create it if needed.

        Returns:
          MQProducer or None:
        """
        if self.__defaultMQProducer is None:
            self.__defaultMQProducer = self.__createProducer()
        return self.__defaultMQProducer

    def __createProducer(self):
        """
        This method is used to create an MQ producer.
        Returns:
          MQProducer or None:
        """
        mqProducer = None
        result = gConfig.getConfigurationTree("/Resources/MQServices")
        if result["OK"]:
            result = createProducer("Monitoring::Queues::%s" % self.__failoverQueueName)
            if not result["OK"]:
                gLogger.debug("Fail to create Producer:", result["Message"])
            else:
                mqProducer = result["Value"]
        else:
            gLogger.warn(
                "No MQ setup for Monitoring in CS. You are running Monitoring without failover:", result["Message"]
            )
        return mqProducer



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-pilot-job-info
# Author :  Philippe Charpentier
########################################################################
"""
Retrieve info about jobs run by the given pilot
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


def _stringInList(subStr, sList):
    resList = []
    for s in sList:
        if subStr.lower() in s.lower():
            resList.append(s)
    return resList


@Script()
def main():
    parameters = ["OwnerDN", "StartExecTime", "EndExecTime"]
    Script.registerSwitch("", "Parameters=", "   List of strings to be matched by job parameters or attributes")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["PilotID:  Grid ID of the pilot"])
    switches, args = Script.parseCommandLine(ignoreErrors=True)
    for switch in switches:
        if switch[0] == "Parameters":
            parameters += [par for par in switch[1].split(",")]
    parameters = [(i, par.lower()) for i, par in enumerate(parameters) if par]

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
    from DIRAC.Interfaces.API.Dirac import Dirac

    diracAdmin = DiracAdmin()
    dirac = Dirac()
    errorList = []

    for gridID in args:

        result = {}
        res = diracAdmin.getPilotInfo(gridID)
        if not res["OK"]:
            errorList.append((gridID, res["Message"]))
        else:
            jobIDs = set([int(jobID) for jobID in res["Value"][gridID]["Jobs"]])
            totCPU = 0
            totWall = 0
            effRequested = False
            for jobID in sorted(jobIDs):
                result.setdefault(jobID, {})
                for func in (dirac.getJobParameters, dirac.getJobAttributes):
                    res = func(jobID)
                    if not res["OK"]:
                        errorList.append(("Job %d" % jobID, res["Message"]))
                    else:
                        params = res["Value"]
                        if "TotalCPUTime(s)" in params:
                            totCPU += float(params["TotalCPUTime(s)"])
                            totWall += float(params["WallClockTime(s)"])
                            params["CPUEfficiency"] = "%s %%" % (
                                100.0 * float(params["TotalCPUTime(s)"]) / float(params["WallClockTime(s)"])
                            )
                        for i, par in parameters:
                            for param in [
                                p
                                for p in _stringInList(str(par), str(params))
                                if not _stringInList(str(p), str(result[jobID]))
                            ]:
                                if param == "CPUEfficiency":
                                    effRequested = True
                                result[jobID]["%d.%s" % (i, param)] = params[param]
                if effRequested:
                    result["CPUEfficiency"] = "%s %%" % (100.0 * totCPU / totWall)
            print(diracAdmin.pPrint.pformat({gridID: result}))

    for error in errorList:
        print("ERROR %s: %s" % error)


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from diraccfg import CFG
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import List


def loadJDLAsCFG(jdl):
    """
    Load a JDL as CFG
    """

    def cleanValue(value):
        value = value.strip()
        if value[0] == '"':
            entries = []
            iPos = 1
            current = ""
            state = "in"
            while iPos < len(value):
                if value[iPos] == '"':
                    if state == "in":
                        entries.append(current)
                        current = ""
                        state = "out"
                    elif state == "out":
                        current = current.strip()
                        if current not in (",",):
                            return S_ERROR("value seems a list but is not separated in commas")
                        current = ""
                        state = "in"
                else:
                    current += value[iPos]
                iPos += 1
            if state == "in":
                return S_ERROR('value is opened with " but is not closed')
            return S_OK(", ".join(entries))
        else:
            return S_OK(value.replace('"', ""))

    def assignValue(key, value, cfg):
        key = key.strip()
        if len(key) == 0:
            return S_ERROR("Invalid key name")
        value = value.strip()
        if not value:
            return S_ERROR("No value for key %s" % key)
        if value[0] == "{":
            if value[-1] != "}":
                return S_ERROR("Value '%s' seems a list but does not end in '}'" % (value))
            valList = List.fromChar(value[1:-1])
            for i in range(len(valList)):
                result = cleanValue(valList[i])
                if not result["OK"]:
                    return S_ERROR("Var %s : %s" % (key, result["Message"]))
                valList[i] = result["Value"]
                if valList[i] is None:
                    return S_ERROR("List value '%s' seems invalid for item %s" % (value, i))
            value = ", ".join(valList)
        else:
            result = cleanValue(value)
            if not result["OK"]:
                return S_ERROR("Var %s : %s" % (key, result["Message"]))
            nV = result["Value"]
            if nV is None:
                return S_ERROR("Value '%s seems invalid" % (value))
            value = nV
        cfg.setOption(key, value)
        return S_OK()

    if jdl[0] == "[":
        iPos = 1
    else:
        iPos = 0
    key = ""
    value = ""
    action = "key"
    insideLiteral = False
    cfg = CFG()
    while iPos < len(jdl):
        char = jdl[iPos]
        if char == ";" and not insideLiteral:
            if key.strip():
                result = assignValue(key, value, cfg)
                if not result["OK"]:
                    return result
            key = ""
            value = ""
            action = "key"
        elif char == "[" and not insideLiteral:
            key = key.strip()
            if not key:
                return S_ERROR("Invalid key in JDL")
            if value.strip():
                return S_ERROR("Key %s seems to have a value and open a sub JDL at the same time" % key)
            result = loadJDLAsCFG(jdl[iPos:])
            if not result["OK"]:
                return result
            subCfg, subPos = result["Value"]
            cfg.createNewSection(key, contents=subCfg)
            key = ""
            value = ""
            action = "key"
            insideLiteral = False
            iPos += subPos
        elif char == "=" and not insideLiteral:
            if action == "key":
                action = "value"
                insideLiteral = False
            else:
                value += char
        elif char == "]" and not insideLiteral:
            key = key.strip()
            if len(key) > 0:
                result = assignValue(key, value, cfg)
                if not result["OK"]:
                    return result
            return S_OK((cfg, iPos))
        else:
            if action == "key":
                key += char
            else:
                value += char
                if char == '"':
                    insideLiteral = not insideLiteral
        iPos += 1

    return S_OK((cfg, iPos))


def dumpCFGAsJDL(cfg, level=1, tab="  "):
    indent = tab * level
    contents = ["%s[" % (tab * (level - 1))]
    sections = cfg.listSections()

    for key in cfg:
        if key in sections:
            contents.append("%s%s =" % (indent, key))
            contents.append("%s;" % dumpCFGAsJDL(cfg[key], level + 1, tab))
        else:
            val = List.fromChar(cfg[key])
            # Some attributes are never lists
            if len(val) < 2 or key in ["Arguments", "Executable", "StdOutput", "StdError"]:
                value = cfg[key]
                try:
                    try_value = float(value)
                    contents.append("%s%s = %s;" % (tab * level, key, value))
                except Exception:
                    contents.append('%s%s = "%s";' % (tab * level, key, value))
            else:
                contents.append("%s%s =" % (indent, key))
                contents.append("%s{" % indent)
                for iPos in range(len(val)):
                    try:
                        value = float(val[iPos])
                    except Exception:
                        val[iPos] = '"%s"' % val[iPos]
                contents.append(",\n".join(["%s%s" % (tab * (level + 1), value) for value in val]))
                contents.append("%s};" % indent)
    contents.append("%s]" % (tab * (level - 1)))
    return "\n".join(contents)



""" pytest for TransformationClient
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access,missing-docstring,invalid-name

import pytest

# sut
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


tc = TransformationClient()


@pytest.mark.parametrize(
    "tsFiles, dictOfNewLFNsStatus, force, expected",
    [
        ({}, {}, False, {}),
        ({}, {"foo": ["status", 2, 1234]}, False, {}),
        ({"foo": ["status", 2, 1234]}, {"foo": "status"}, False, {}),
        ({"foo": ["status", 2, 1234]}, {"foo": "statusA"}, False, {"foo": "statusA"}),
        ({"foo": ["status", 2, 1234], "bar": ["status", 2, 5678]}, {"foo": "status"}, False, {}),
        ({"foo": ["status", 2, 1234], "bar": ["status", 2, 5678]}, {"foo": "statusA"}, False, {"foo": "statusA"}),
        (
            {"foo": ["status", 2, 1234], "bar": ["status", 2, 5678]},
            {"foo": "A", "bar": "B"},
            False,
            {"foo": "A", "bar": "B"},
        ),
        ({"foo": ["status", 2, 1234]}, {"foo": "A", "bar": "B"}, False, {"foo": "A"}),
        ({"foo": ["Assigned", 2, 1234]}, {"foo": "A", "bar": "B"}, False, {"foo": "A"}),
        (
            {"foo": ["Assigned", 2, 1234], "bar": ["Assigned", 2, 5678]},
            {"foo": "Assigned", "bar": "Processed"},
            False,
            {"bar": "Processed"},
        ),
        (
            {"foo": ["Processed", 2, 1234], "bar": ["Unused", 2, 5678]},
            {"foo": "Assigned", "bar": "Processed"},
            False,
            {"bar": "Processed"},
        ),
        (
            {"foo": ["Processed", 2, 1234], "bar": ["Unused", 2, 5678]},
            {"foo": "Assigned", "bar": "Processed"},
            True,
            {"foo": "Assigned", "bar": "Processed"},
        ),
        (
            {"foo": ["MaxReset", 12, 1234], "bar": ["Processed", 22, 5678]},
            {"foo": "Unused", "bar": "Unused"},
            False,
            {},
        ),
        (
            {"foo": ["MaxReset", 12, 1234], "bar": ["Processed", 22, 5678]},
            {"foo": "Unused", "bar": "Unused"},
            True,
            {"foo": "Unused", "bar": "Unused"},
        ),
        (
            {"foo": ["Assigned", 20, 1234], "bar": ["Processed", 2, 5678]},
            {"foo": "Unused", "bar": "Unused"},
            False,
            {"foo": "MaxReset"},
        ),
    ],
)
def test__applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, force, expected):
    res = tc._applyTransformationFilesStateMachine(tsFiles, dictOfNewLFNsStatus, force)
    assert res == expected



"""
Logging
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import logging
import os

from DIRAC import S_ERROR
from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels
from DIRAC.Core.Utilities.LockRing import LockRing


class Logging(object):
    """
    - Logging is a wrapper of the logger object from the standard "logging" library which integrates some DIRAC concepts.
    - It aimed at seamlessly replacing the previous gLogger implementation and thus provides the same interface.
    - Logging is generally used to create log records, that are then sent to pre-determined backends.

    Each Logging embeds a logger of "logging". Logging can instanciate "children" logging objects and
    all Logging objects inherit from the configuration of LoggingRoot, the first Logging object to be instanciated.
    """

    # componentName is a class variable: the component name is the same for every Logging objects
    # its default value is "Framework" but it can be configured in initialize() in LoggingRoot
    # it can be composed by the system name and the component name. For instance: "Monitoring/Atom"
    _componentName = "Framework"
    # use the lockRing singleton to save the Logging object
    _lockRing = LockRing()
    # lock the configuration of the Logging
    _lockConfig = _lockRing.getLock("config")

    def __init__(self, father=None, fatherName="", name="", customName=""):
        """
        Initialization of the Logging object. By default, 'fatherName' and 'name' are empty,
        because getChild only accepts string and the first empty string corresponds to the root logger.
        Example:
        >>> logging.getLogger('') == logging.getLogger('root') # root logger
        >>> logging.getLogger('root').getChild('log') == logging.getLogger('log') # log child of root

        :param Logging father: father of this new Logging.
        :param str fatherName: name of the father logger in the chain.
        :param str name: name of the logger in the chain.
        :param str customName: name of the logger in the chain:
                                - "root" does not appear at the beginning of the chain
                                - hierarchy "." are replaced by "\"
        """

        # Logging chain
        self._children = {}
        self._parent = father

        # initialize display options and level with the ones of the Logging parent
        if self._parent is not None:
            self._options = self._parent.getDisplayOptions()
        else:
            self._options = {
                "headerIsShown": True,
                "timeStampIsShown": True,
                "contextIsShown": True,
                "threadIDIsShown": False,
                "color": False,
            }

        # dictionary of the options modifications: give the same behaviour that the "logging" level
        # - propagation from the parent to the children when their levels are not set by the developer
        # - stop the propagation when a developer set a level to a child
        self._optionsModified = {
            "headerIsShown": False,
            "timeStampIsShown": False,
            "contextIsShown": False,
            "threadIDIsShown": False,
        }

        self._backendsList = []

        # name of the Logging
        self.name = str(name)
        self._logger = logging.getLogger(fatherName).getChild(self.name)

        # update the custom name of the Logging adding the new Logging name in the entire path
        self._customName = os.path.join("/", customName, self.name)

        # Locks to make Logging thread-safe
        # we use RLock to prevent blocking in the Logging
        # lockInit to protect the initialization of a sublogger
        self._lockInit = self._lockRing.getLock("init")
        # lockOptions to protect the option modifications and the backendsList
        self._lockOptions = self._lockRing.getLock("options", recursive=True)
        # lockLevel to protect the level
        self._lockLevel = self._lockRing.getLock("level", recursive=True)
        # lockObjectLoader to protect the ObjectLoader singleton
        self._lockObjectLoader = self._lockRing.getLock("objectLoader")

    def showHeaders(self, yesno=True):
        """
        Depending on the value, display or not the prefix of the message.

        :param bool yesno: determine the log record format
        """
        self._setOption("headerIsShown", yesno)

    def showThreadIDs(self, yesno=True):
        """
        Depending on the value, display or not the thread ID.
        Make sure to enable the headers: showHeaders(True) before

        :param bool yesno: determe the log record format
        """
        self._setOption("threadIDIsShown", yesno)

    def showTimeStamps(self, yesno=True):
        """
        Depending on the value, display or not the timestamp of the message.
        Make sure to enable the headers: showHeaders(True) before

        :param bool yesno: determine the log record format
        """
        self._setOption("timeStampIsShown", yesno)

    def showContexts(self, yesno=True):
        """
        Depending on the value, display or not the context of the message.
        Make sure to enable the headers: showHeaders(True) before

        :param bool yesno: determine the log record format
        """
        self._setOption("contextIsShown", yesno)

    def _setOption(self, optionName, value, directCall=True):
        """
        Depending on the value, modify the value of the option and propagate the option to the children.
        The options of the children will be updated if they were not modified before by a developer.

        :param str optionName: name of the option to modify
        :param bool value: value of the option to set
        :param bool directCall: indicate whether the call is performed by a developer
        """
        # lock to prevent that two threads change the options at the same time
        self._lockOptions.acquire()
        try:
            if self._optionsModified[optionName] and not directCall:
                return

            if directCall:
                self._optionsModified[optionName] = True

            # update option
            self._options[optionName] = value

            # propagate in the children
            for child in self._children.values():
                child._setOption(optionName, value, directCall=False)  # pylint: disable=protected-access
        finally:
            self._lockOptions.release()

    def registerBackends(self, desiredBackends, backendOptions=None):
        """
        Attach a list of backends to the Logging object.
        Convert backend names to backend class names to Backend objects and add them to the Logging object

        :param desiredBackends: list of different names attaching to differents backends.
                                 list of the possible values: ['stdout', 'stderr', 'file']
        :param backendOptions: dictionary of different backend options. Example: FileName='/tmp/log.txt'
        """
        for backendName in desiredBackends:
            self.registerBackend(backendName, backendOptions)

    def registerBackend(self, desiredBackend, backendOptions=None):
        """
        Attach a backend to the Logging object.
        Convert backend name to backend class name to a Backend object and add it to the Logging object

        :param desiredBackend: a name attaching to a backend type. List of possible values: ['stdout', 'stderr', 'file']
        :param backendOptions: dictionary of different backend options. Example: FileName='/tmp/log.txt'
        """
        # Remove white space and capitalize the first letter
        desiredBackend = desiredBackend.strip()
        desiredBackend = desiredBackend[0].upper() + desiredBackend[1:]
        _class = self.__loadLogClass("Resources.LogBackends.%sBackend" % desiredBackend)
        if _class["OK"]:
            # add the backend instance to the Logging
            self._addBackend(_class["Value"], backendOptions)
        else:
            self.warn("%s is not a valid backend name." % desiredBackend)

    def _addBackend(self, backendType, backendOptions=None):
        """
        Attach a Backend object to the Logging object.

        :param Backend backend: Backend object that has to be added
        :param backendOptions: a dictionary of different backend options. Example: {'FileName': '/tmp/log.txt'}
        """
        # lock to prevent that the level change before adding the new backend in the backendsList
        # and to prevent a change of the backendsList during the reading of the
        # list
        self._lockLevel.acquire()
        self._lockOptions.acquire()
        try:
            backend = backendType(backendOptions)
            self._logger.addHandler(backend.getHandler())
            self._addFilter(backend, backendOptions)
            self._backendsList.append(backend)
        finally:
            self._lockLevel.release()
            self._lockOptions.release()

    def _addFilter(self, backend, backendOptions):
        """
        Create a filter and add it to the handler of the backend.
        """
        for filterName in self.__getFilterList(backendOptions):
            options = self.__getFilterOptionsFromCFG(filterName)
            _class = self.__loadLogClass("Resources.LogFilters.%s" % options.get("Plugin"))
            if _class["OK"]:
                # add the backend instance to the Logging
                backend.getHandler().addFilter(_class["Value"](options))
            else:
                self.warn("%r is not a valid Filter name." % filterName)

    def __getFilterList(self, backendOptions):
        """
        Return list of defined filters.
        """
        if not (isinstance(backendOptions, dict) and "Filter" in backendOptions):
            return []
        return [fil.strip() for fil in backendOptions["Filter"].split(",") if fil.strip()]

    def __getFilterOptionsFromCFG(self, logFilter):
        """Get filter options from the configuration..

        :param logFilter: filter identifier: stdout, file, f04
        """
        # We have to put the import lines here to avoid a dependancy loop
        from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getFilterConfig

        # Search filters config in the resources section
        retDictRessources = getFilterConfig(logFilter)
        if retDictRessources["OK"]:
            return retDictRessources["Value"]
        return {}

    def setLevel(self, levelName):
        """
        Check if the level name exists and set it.

        :param levelName: string representing the level to give to the logger
        :return: boolean representing if the setting is done or not
        """
        result = False
        if levelName.upper() in LogLevels.getLevelNames():
            self._logger.setLevel(LogLevels.getLevelValue(levelName))
            result = True
        return result

    def getLevel(self):
        """
        :return: the name of the level
        """
        return LogLevels.getLevel(self._logger.getEffectiveLevel())

    def shown(self, levelName):
        """
        Determine whether messages with a certain level will be displayed.

        :param levelName: string representing the level to analyse

        :return: boolean which give the answer
        """
        # lock to prevent a level change
        self._lockLevel.acquire()
        try:
            result = False
            if levelName.upper() in LogLevels.getLevelNames():
                result = LogLevels.getLevelValue(self.getLevel()) <= LogLevels.getLevelValue(levelName)
            return result
        finally:
            self._lockLevel.release()

    @classmethod
    def getName(cls):
        """
        :return: "system name/component name"
        """
        return cls._componentName

    def getSubName(self):
        """
        :return: the name of the logger
        """
        return self.name

    def getDisplayOptions(self):
        """
        :return: the dictionary of the display options and their values. Must not be redefined
        """
        # lock to save the options which can be modified
        self._lockOptions.acquire()
        try:
            # copy the dictionary to avoid that every Logging has the same
            options = self._options.copy()
            return options
        finally:
            self._lockOptions.release()

    def __loadLogClass(self, modulePath):
        """Load class thread-safe."""
        # import ObjectLoader here to avoid a dependancy loop
        from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

        objLoader = ObjectLoader()
        # lock to avoid problem in ObjectLoader which is a singleton not
        # thread-safe
        self._lockObjectLoader.acquire()
        try:
            # load the Backend class
            return objLoader.loadObject(modulePath)
        finally:
            self._lockObjectLoader.release()
        return S_ERROR()

    @staticmethod
    def getAllPossibleLevels():
        """
        :return: a list of all levels available
        """
        return LogLevels.getLevelNames()

    def always(self, sMsg, sVarMsg=""):
        """
        Always level
        """
        return self._createLogRecord(LogLevels.ALWAYS, sMsg, sVarMsg)

    def notice(self, sMsg, sVarMsg=""):
        """
        Notice level
        """
        return self._createLogRecord(LogLevels.NOTICE, sMsg, sVarMsg)

    def info(self, sMsg, sVarMsg=""):
        """
        Info level
        """
        return self._createLogRecord(LogLevels.INFO, sMsg, sVarMsg)

    def verbose(self, sMsg, sVarMsg=""):
        """
        Verbose level
        """
        return self._createLogRecord(LogLevels.VERBOSE, sMsg, sVarMsg)

    def debug(self, sMsg, sVarMsg=""):
        """
        Debug level
        """
        return self._createLogRecord(LogLevels.DEBUG, sMsg, sVarMsg)

    def warn(self, sMsg, sVarMsg=""):
        """
        Warn
        """
        return self._createLogRecord(LogLevels.WARN, sMsg, sVarMsg)

    def error(self, sMsg, sVarMsg=""):
        """
        Error level
        """
        return self._createLogRecord(LogLevels.ERROR, sMsg, sVarMsg)

    def exception(self, sMsg="", sVarMsg="", lException=False, lExcInfo=False):
        """
        Exception level
        """
        _ = lException  # Make pylint happy
        _ = lExcInfo
        return self._createLogRecord(LogLevels.ERROR, sMsg, sVarMsg, exc_info=True)

    def fatal(self, sMsg, sVarMsg=""):
        """
        Fatal level
        """
        return self._createLogRecord(LogLevels.FATAL, sMsg, sVarMsg)

    def _createLogRecord(self, level, sMsg, sVarMsg, exc_info=False, local_context=None):
        """
        Create a log record according to the level of the message.

        - The log record is always sent to the different backends
        - Backends have their own levels and may manage the display of the log record

        :param int level: level of the log record
        :param str sMsg: message
        :param str sVarMsg: additional message
        :param bool exc_info: indicates whether the stacktrace has to appear in the log record
        :param dict local_context: Extra information propagated as extra to the formater.
                                   It is meant to be used only by the LocalSubLogger

        :return: boolean representing the result of the log record creation
        """

        # lock to prevent a level change after that the log is sent.
        self._lockLevel.acquire()
        try:
            # exc_info is only for exception to add the stack trace

            # extra is a way to add extra attributes to the log record:
            # - 'componentname': the system/component name
            # - 'varmessage': the variable message
            # - 'customname' : the name of the logger for the DIRAC usage: without 'root' and separated with '/'
            # as log records, extras attributes are not camel case
            extra = {
                "componentname": self._componentName,
                "varmessage": str(sVarMsg),
                "spacer": "" if not sVarMsg else " ",
                "customname": self._customName,
            }

            # options such as headers and threadIDs also depend on the logger, we have to add them to extra
            extra.update(self._options)

            # This typically contains local custom names
            if local_context:
                extra.update(local_context)

            self._logger.log(level, "%s", sMsg, exc_info=exc_info, extra=extra)
            # check whether the message is displayed
            isSent = LogLevels.getLevelValue(self.getLevel()) <= level
            return isSent
        finally:
            self._lockLevel.release()

    def showStack(self):
        """
        Display a debug message without any content.

        :return: boolean, True if the message is sent, else False
        """
        return self.debug("")

    def getSubLogger(self, subName, child=True):
        """
        Create a new Logging object, child of this Logging, if it does not exists.

        .. warning::

           For very short lived sub logger, consider :py:meth:`getLocalSubLogger` instead

        :param str subName: name of the child Logging
        """
        _ = child  # make pylint happy
        # lock to prevent that the method initializes two Logging for the same 'logging' logger
        # and to erase the existing _children[subName]
        self._lockInit.acquire()
        try:
            # Check if the object has a child with "subName".
            result = self._children.get(subName)
            if result is not None:
                return result
            # create a new child Logging
            childLogging = Logging(self, self._logger.name, subName, self._customName)

            self._children[subName] = childLogging
            return childLogging
        finally:
            self._lockInit.release()

    class LocalSubLogger(object):
        """
        This is inspired from the standard LoggingAdapter.
        The idea is to provide an interface which looks like a Logger,
        but does not implement all the features.
        You can basically just create it, and log messages.
        You cannot create subLogger from it

        This is to be used for very short lived sub logger. It allows to
        give context information (like a jobID) without creating a new logging
        object, which ends up eating all the memory
        (see https://github.com/DIRACGrid/DIRAC/issues/5280)
        """

        def __init__(self, logger, extra):
            """
            :param logger: :py:class:`Logging` object on which to be based
            :param extra: dictionary of extra information to be passed
            """

            self.logger = logger
            self.extra = extra

        def always(self, sMsg, sVarMsg=""):
            """
            Always level
            """
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.ALWAYS, sMsg, sVarMsg, local_context=self.extra
            )

        def notice(self, sMsg, sVarMsg=""):
            """
            Notice level
            """
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.NOTICE, sMsg, sVarMsg, local_context=self.extra
            )

        def info(self, sMsg, sVarMsg=""):
            """
            Info level
            """
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.INFO, sMsg, sVarMsg, local_context=self.extra
            )

        def verbose(self, sMsg, sVarMsg=""):
            """
            Verbose level
            """
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.VERBOSE, sMsg, sVarMsg, local_context=self.extra
            )

        def debug(self, sMsg, sVarMsg=""):
            """
            Debug level
            """
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.DEBUG, sMsg, sVarMsg, local_context=self.extra
            )

        def warn(self, sMsg, sVarMsg=""):
            """
            Warn
            """
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.WARN, sMsg, sVarMsg, local_context=self.extra
            )

        def error(self, sMsg, sVarMsg=""):
            """
            Error level
            """
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.ERROR, sMsg, sVarMsg, local_context=self.extra
            )

        def exception(self, sMsg="", sVarMsg="", lException=False, lExcInfo=False):
            """
            Exception level
            """
            _ = lException  # Make pylint happy
            _ = lExcInfo
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.ERROR, sMsg, sVarMsg, exc_info=True, local_context=self.extra
            )

        def fatal(self, sMsg, sVarMsg=""):
            """
            Fatal level
            """
            return self.logger._createLogRecord(  # pylint: disable=protected-access
                LogLevels.FATAL, sMsg, sVarMsg, local_context=self.extra
            )

    def getLocalSubLogger(self, subName):
        """
        Create a subLogger which is meant to have very short lifetime,
        (e.g. when you want to add the jobID in the name)

        .. warning::
          This is a light version of a logger, read the documentation of
          :py:class:`LocalSubLogger` carefully

        :param str subName: name of the child Logging
        """

        return Logging.LocalSubLogger(self, dict(local_name=subName))



#!/usr/bin/env python
"""
Script to update pilot version in CS
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch(
        "v:",
        "vo=",
        "Location of pilot version in CS /Operations/<vo>/Pilot/Version"
        " (default value specified in CS under /DIRAC/DefaultSetup)",
    )
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("version: pilot version you want to update to")
    Script.parseCommandLine(ignoreErrors=False)

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    version = Script.getPositionalArgs(group=True)

    vo = None
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "v" or switch[0] == "vo":
            vo = switch[1]

    from DIRAC import S_OK, S_ERROR
    from DIRAC import gConfig, gLogger
    from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI

    def updatePilot(version, vo):
        """
        Update in the CS the pilot version used,
        If only one version present in CS it's overwritten.
        If two versions present, the new one is added and the last removed

        :param version: version vArBpC of pilot you want to use
        :param vo: Location of pilot version in CS /Operations/<vo>/Pilot/Version
        """
        setup = vo
        if not vo:
            setup = gConfig.getValue("/DIRAC/DefaultSetup")
        if not setup:
            return S_ERROR("No value set for /DIRAC/DefaultSetup in CS")

        pilotVersion = gConfig.getValue("Operations/%s/Pilot/Version" % setup, [])
        if not pilotVersion:
            return S_ERROR("No pilot version set under Operations/%s/Pilot/Version in CS" % setup)

        pilotVersion.pop()
        pilotVersion.insert(0, version)
        api = CSAPI()
        api.setOption("Operations/%s/Pilot/Version" % setup, ", ".join(pilotVersion))
        result = api.commit()
        if not result["OK"]:
            gLogger.fatal("Could not commit new version of pilot!")
            return result

        newVersion = gConfig.getValue("Operations/%s/Pilot/Version" % setup)
        return S_OK("New version of pilot set to %s" % newVersion)

    result = updatePilot(version, vo)
    if not result["OK"]:
        gLogger.fatal(result["Message"])
        DIRAC.exit(1)
    gLogger.notice(result["Value"])
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :   tornado-start-CS
# Author : Louis MARTIN
########################################################################
# Just run this script to start Tornado and CS service
# Use dirac.cfg (or other cfg given in the command line) to change port

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import sys
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():

    if os.environ.get("DIRAC_USE_TORNADO_IOLOOP", "false").lower() not in ("yes", "true"):
        raise RuntimeError(
            "DIRAC_USE_TORNADO_IOLOOP is not defined in the environment."
            + "\n"
            + "It is necessary to run with Tornado."
            + "\n"
            + "https://dirac.readthedocs.io/en/latest/DeveloperGuide/TornadoServices/index.html"
        )

    from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceSection
    from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
    from DIRAC.ConfigurationSystem.private.Refresher import gRefresher
    from DIRAC.Core.Utilities.DErrno import includeExtensionErrors
    from DIRAC.Core.Tornado.Server.TornadoServer import TornadoServer
    from DIRAC.FrameworkSystem.Client.Logger import gLogger

    if gConfigurationData.isMaster():
        gRefresher.disable()

    localCfg = Script.localCfg
    localCfg.addMandatoryEntry("/DIRAC/Setup")
    localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "yes")
    localCfg.addDefaultEntry("LogLevel", "INFO")
    localCfg.addDefaultEntry("LogColor", True)
    resultDict = localCfg.loadUserData()
    if not resultDict["OK"]:
        gLogger.initialize("Tornado-CS", "/")
        gLogger.error("There were errors when loading configuration", resultDict["Message"])
        sys.exit(1)

    includeExtensionErrors()

    gLogger.initialize("Tornado-CS", "/")

    # get the specific master CS port
    try:
        csPort = int(gConfigurationData.extractOptionFromCFG("%s/Port" % getServiceSection("Configuration/Server")))
    except TypeError:
        csPort = None

    serverToLaunch = TornadoServer(services="Configuration/Server", port=csPort)
    serverToLaunch.startTornado()


if __name__ == "__main__":
    main()



""" FreeDiskSpaceCommand
    The Command gets the free space that is left in a Storage Element

    Note: there are, still, many references to "space tokens",
    for example ResourceManagementClient().selectSpaceTokenOccupancyCache(token=elementName)
    This is for historical reasons, and shoud be fixed one day.
    For the moment, when you see "token" or "space token" here, just read "StorageElement".

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys
import errno

from datetime import datetime

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.File import convertSizeUnits
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from DIRAC.AccountingSystem.Client.Types.StorageOccupancy import StorageOccupancy
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.ResourceStatusSystem.Command.Command import Command
from DIRAC.ResourceStatusSystem.Utilities import CSHelpers
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient


class FreeDiskSpaceCommand(Command):
    """
    Uses diskSpace method to get the free space
    """

    def __init__(self, args=None, clients=None):

        super(FreeDiskSpaceCommand, self).__init__(args, clients=clients)

        self.rmClient = ResourceManagementClient()

    def _prepareCommand(self):
        """
        FreeDiskSpaceCommand requires one argument:
        - name : <str>
        """

        if "name" not in self.args:
            return S_ERROR('"name" not found in self.args')
        elementName = self.args["name"]

        # We keep TB as default as this is what was used (and will still be used)
        # in the policy for "space tokens" ("real", "data" SEs)
        unit = self.args.get("unit", "TB")

        return S_OK((elementName, unit))

    def doNew(self, masterParams=None):
        """
        Gets the parameters to run, either from the master method or from its
        own arguments.

        Gets the total and the free disk space of a storage element
        and inserts the results in the SpaceTokenOccupancyCache table
        of ResourceManagementDB database.

        The result is also returned to the caller, not only inserted.
        What is inserted in the DB will normally be in MB,
        what is returned will be in the specified unit.
        """

        if masterParams is not None:
            elementName, unit = masterParams
        else:
            params = self._prepareCommand()
            if not params["OK"]:
                return params
            elementName, unit = params["Value"]

        se = StorageElement(elementName)
        occupancyResult = se.getOccupancy(unit=unit)
        if not occupancyResult["OK"]:
            return occupancyResult
        occupancy = occupancyResult["Value"]
        free = occupancy["Free"]
        total = occupancy["Total"]

        endpointResult = CSHelpers.getStorageElementEndpoint(elementName)
        if not endpointResult["OK"]:
            return endpointResult
        # We only take the first endpoint, in case there are severals of them (which is normal).
        # Most probably not ideal, because it would be nice to stay consistent, but well...
        endpoint = endpointResult["Value"][0]

        results = {"Endpoint": endpoint, "Free": free, "Total": total, "ElementName": elementName}
        result = self._storeCommand(results)
        if not result["OK"]:
            return result

        return S_OK({"Free": free, "Total": total})

    def _storeCommand(self, results):
        """
        Stores the results in the cache (SpaceTokenOccupancyCache),
        and adds records to the StorageOccupancy accounting.

        :param dict results: something like {'ElementName': 'CERN-HIST-EOS',
                                             'Endpoint': 'httpg://srm-eoslhcb-bis.cern.ch:8443/srm/v2/server',
                                             'Free': 3264963586.10073,
                                             'Total': 8000000000.0}
        :returns: S_OK/S_ERROR dict
        """

        # Stores in cache
        res = self.rmClient.addOrModifySpaceTokenOccupancyCache(
            endpoint=results["Endpoint"],
            lastCheckTime=datetime.utcnow(),
            free=results["Free"],
            total=results["Total"],
            token=results["ElementName"],
        )
        if not res["OK"]:
            self.log.error("Error calling addOrModifySpaceTokenOccupancyCache", res["Message"])
            return res

        # Now proceed with the accounting
        siteRes = DMSHelpers().getLocalSiteForSE(results["ElementName"])
        if not siteRes["OK"]:
            return siteRes

        accountingDict = {
            "StorageElement": results["ElementName"],
            "Endpoint": results["Endpoint"],
            "Site": siteRes["Value"] if siteRes["Value"] else "unassigned",
        }

        results["Used"] = results["Total"] - results["Free"]

        for sType in ["Total", "Free", "Used"]:
            spaceTokenAccounting = StorageOccupancy()
            spaceTokenAccounting.setNowAsStartAndEndTime()
            spaceTokenAccounting.setValuesFromDict(accountingDict)
            spaceTokenAccounting.setValueByKey("SpaceType", sType)
            spaceTokenAccounting.setValueByKey("Space", int(convertSizeUnits(results[sType], "MB", "B")))

            res = gDataStoreClient.addRegister(spaceTokenAccounting)
            if not res["OK"]:
                self.log.warn("Could not commit register", res["Message"])
                continue

        return gDataStoreClient.commit()

    def doCache(self):
        """
        This is a method that gets the element's details from the spaceTokenOccupancyCache DB table.
        It will return a dictionary with th results, converted to "correct" unit.
        """

        params = self._prepareCommand()
        if not params["OK"]:
            return params
        elementName, unit = params["Value"]

        result = self.rmClient.selectSpaceTokenOccupancyCache(token=elementName)

        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR(errno.ENODATA, "No occupancy recorded")

        # results are normally in 'MB'
        free = result["Value"][0][3]
        total = result["Value"][0][4]

        free = convertSizeUnits(free, "MB", unit)
        total = convertSizeUnits(total, "MB", unit)

        if free == -sys.maxsize or total == -sys.maxsize:
            return S_ERROR("No valid unit specified")

        return S_OK({"Free": free, "Total": total})

    def doMaster(self):
        """
        This method calls the doNew method for each storage element
        that exists in the CS.
        """

        for name in DMSHelpers().getStorageElements():
            try:
                # keeping TB as default
                diskSpace = self.doNew((name, "MB"))
                if not diskSpace["OK"]:
                    self.log.warn("Unable to calculate free/total disk space", "name: %s" % name)
                    self.log.warn(diskSpace["Message"])
                    continue
            except Exception as excp:  # pylint: disable=broad-except
                self.log.error("Failed to get SE FreeDiskSpace information ==> SE skipped", name)
                self.log.exception("Operation finished with exception: ", lException=excp)

        # Clear the cache
        return self._cleanCommand()

    def _cleanCommand(self, toDelete=None):
        """Clean the spaceTokenOccupancy table from old endpoints

        :param tuple toDelete: endpoint to remove (endpoint, storage_element_name),
                               e.g. ('httpg://srm-lhcb.cern.ch:8443/srm/managerv2', CERN-RAW)
        """
        if not toDelete:
            toDelete = []

            res = self.rmClient.selectSpaceTokenOccupancyCache()
            if not res["OK"]:
                return res
            storedSEsSet = set([(sse[0], sse[1]) for sse in res["Value"]])

            currentSEsSet = set()
            currentSEs = DMSHelpers().getStorageElements()
            for cse in currentSEs:
                res = CSHelpers.getStorageElementEndpoint(cse)
                if not res["OK"]:
                    self.log.warn("Could not get endpoint", res["Message"])
                    continue
                endpoint = res["Value"][0]

                currentSEsSet.add((endpoint, cse))
            toDelete = list(storedSEsSet - currentSEsSet)

        else:
            toDelete = [toDelete]

        for ep in toDelete:
            res = self.rmClient.deleteSpaceTokenOccupancyCache(ep[0], ep[1])
            if not res["OK"]:
                self.log.warn("Could not delete entry from SpaceTokenOccupancyCache", res["Message"])

        return S_OK()



""" Few unit tests for LCG clients
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import mock

from datetime import datetime, timedelta
from DIRAC.Core.LCG.GOCDBClient import GOCDBClient


mockRSS = mock.MagicMock()
GOCCli = GOCDBClient()

# data

now = datetime.utcnow().replace(microsecond=0, second=0)
tomorrow = datetime.utcnow().replace(microsecond=0, second=0) + timedelta(hours=24)
inAWeek = datetime.utcnow().replace(microsecond=0, second=0) + timedelta(days=7)
nowLess12h = str(now - timedelta(hours=12))[:-3]
nowPlus8h = str(now + timedelta(hours=8))[:-3]
nowPlus24h = str(now + timedelta(hours=24))[:-3]
nowPlus40h = str(now + timedelta(hours=40))[:-3]
nowPlus50h = str(now + timedelta(hours=50))[:-3]
nowPlus60h = str(now + timedelta(hours=60))[:-3]

XML_site_ongoing = '<?xml version="1.0"?>\n<ROOT><DOWNTIME ID="78505456" PRIMARY_KEY="28490G0"'
XML_site_ongoing += ' CLASSIFICATION="SCHEDULED"><SITENAME>GRISU-ENEA-GRID</SITENAME>'
XML_site_ongoing += "<SEVERITY>OUTAGE</SEVERITY>"
XML_site_ongoing += "<DESCRIPTION>Software problems SITE</DESCRIPTION>"
XML_site_ongoing += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_site_ongoing += "<START_DATE>1276360500</START_DATE>"
XML_site_ongoing += "<END_DATE>1276878660</END_DATE>"
XML_site_ongoing += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
XML_site_ongoing += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
XML_site_ongoing += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_site_ongoing += "</DOWNTIME></ROOT>\n"

XML_node_ongoing = '<?xml version="1.0"?>\n<ROOT><DOWNTIME ID="78505455" PRIMARY_KEY="28490G0"'
XML_node_ongoing += ' CLASSIFICATION="SCHEDULED"><HOSTNAME>egse-cresco.portici.enea.it</HOSTNAME>'
XML_node_ongoing += "<HOSTED_BY>GRISU-ENEA-GRID</HOSTED_BY>"
XML_node_ongoing += "<SEVERITY>OUTAGE</SEVERITY>"
XML_node_ongoing += "<DESCRIPTION>Software problems</DESCRIPTION>"
XML_node_ongoing += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_node_ongoing += "<START_DATE>1276360500</START_DATE><END_DATE>1276878660</END_DATE>"
XML_node_ongoing += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
XML_node_ongoing += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
XML_node_ongoing += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_node_ongoing += "</DOWNTIME></ROOT>\n"

xml_endpoint_and_affected_ongoing = '<?xml version="1.0"?>\n<ROOT>'
xml_endpoint_and_affected_ongoing += '<DOWNTIME ID="29118" PRIMARY_KEY="109962G0" CLASSIFICATION="UNSCHEDULED">'
xml_endpoint_and_affected_ongoing += "<PRIMARY_KEY>109962G0</PRIMARY_KEY>"
xml_endpoint_and_affected_ongoing += "<HOSTNAME>lhcbsrm-kit.gridka.de</HOSTNAME>"
xml_endpoint_and_affected_ongoing += "<SERVICE_TYPE>SRM</SERVICE_TYPE>"
xml_endpoint_and_affected_ongoing += "<ENDPOINT>lhcbsrm-kit.gridka.deSRM</ENDPOINT>"
xml_endpoint_and_affected_ongoing += "<HOSTED_BY>FZK-LCG2</HOSTED_BY>"
xml_endpoint_and_affected_ongoing += "<GOCDB_PORTAL_URL>https://goc.egi.eu/bof</GOCDB_PORTAL_URL>"
xml_endpoint_and_affected_ongoing += "<AFFECTED_ENDPOINTS>"
xml_endpoint_and_affected_ongoing += "<ENDPOINT>"
xml_endpoint_and_affected_ongoing += "<ID>7517</ID>"
xml_endpoint_and_affected_ongoing += "<NAME>lhcbsrm-disk-kit</NAME>"
xml_endpoint_and_affected_ongoing += "<URL>lhcbsrm-disk-kit.gridka.de</URL>"
xml_endpoint_and_affected_ongoing += "<INTERFACENAME>SRM</INTERFACENAME>"
xml_endpoint_and_affected_ongoing += "<ENDPOINT_MONITORED>N</ENDPOINT_MONITORED>"
xml_endpoint_and_affected_ongoing += "</ENDPOINT>"
xml_endpoint_and_affected_ongoing += "</AFFECTED_ENDPOINTS>"
xml_endpoint_and_affected_ongoing += "<SEVERITY>OUTAGE</SEVERITY>"
xml_endpoint_and_affected_ongoing += "<DESCRIPTION>Namespace reordering</DESCRIPTION>"
xml_endpoint_and_affected_ongoing += "<INSERT_DATE>1595233003</INSERT_DATE>"
xml_endpoint_and_affected_ongoing += "<START_DATE>1595314800</START_DATE>"
xml_endpoint_and_affected_ongoing += "<END_DATE>1595343600</END_DATE>"
xml_endpoint_and_affected_ongoing += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
xml_endpoint_and_affected_ongoing += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
xml_endpoint_and_affected_ongoing += "</DOWNTIME></ROOT>\n"

xml_endpoint_and_affected_ongoing_broken = '<?xml version="1.0"?>\n<ROOT>'
xml_endpoint_and_affected_ongoing_broken += '<DOWNTIME ID="29118" PRIMARY_KEY="109962G0" CLASSIFICATION="UNSCHEDULED">'
xml_endpoint_and_affected_ongoing_broken += "<PRIMARY_KEY>109962G0</PRIMARY_KEY>"
xml_endpoint_and_affected_ongoing_broken += "<HOSTNAME>lhcbsrm-kit.gridka.de</HOSTNAME>"
xml_endpoint_and_affected_ongoing_broken += "<SERVICE_TYPE>SRM</SERVICE_TYPE>"
xml_endpoint_and_affected_ongoing_broken += "<ENDPOINT>lhcbsrm-kit.gridka.deSRM</ENDPOINT>"
xml_endpoint_and_affected_ongoing_broken += "<HOSTED_BY>FZK-LCG2</HOSTED_BY>"
xml_endpoint_and_affected_ongoing_broken += "<GOCDB_PORTAL_URL>https://goc.egi.eu/bof</GOCDB_PORTAL_URL>"
xml_endpoint_and_affected_ongoing_broken += "<AFFECTED_ENDPOINTS>"
xml_endpoint_and_affected_ongoing_broken += "<ENDPOINT>"
xml_endpoint_and_affected_ongoing_broken += "<ID>7517</ID>"
xml_endpoint_and_affected_ongoing_broken += "<NAME>lhcbsrm-disk-kit</NAME>"
xml_endpoint_and_affected_ongoing_broken += "<INTERFACENAME>SRM</INTERFACENAME>"
xml_endpoint_and_affected_ongoing_broken += "<ENDPOINT_MONITORED>N</ENDPOINT_MONITORED>"
xml_endpoint_and_affected_ongoing_broken += "</ENDPOINT>"
xml_endpoint_and_affected_ongoing_broken += "</AFFECTED_ENDPOINTS>"
xml_endpoint_and_affected_ongoing_broken += "<SEVERITY>OUTAGE</SEVERITY>"
xml_endpoint_and_affected_ongoing_broken += "<DESCRIPTION>Namespace reordering</DESCRIPTION>"
xml_endpoint_and_affected_ongoing_broken += "<INSERT_DATE>1595233003</INSERT_DATE>"
xml_endpoint_and_affected_ongoing_broken += "<START_DATE>1595314800</START_DATE>"
xml_endpoint_and_affected_ongoing_broken += "<END_DATE>1595343600</END_DATE>"
xml_endpoint_and_affected_ongoing_broken += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
xml_endpoint_and_affected_ongoing_broken += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
xml_endpoint_and_affected_ongoing_broken += "</DOWNTIME></ROOT>\n"

xml_endpoint_and_affected_ongoing_no_url = '<?xml version="1.0"?>\n<ROOT>'
xml_endpoint_and_affected_ongoing_no_url += '<DOWNTIME ID="29118" PRIMARY_KEY="109962G0" CLASSIFICATION="UNSCHEDULED">'
xml_endpoint_and_affected_ongoing_no_url += "<PRIMARY_KEY>109962G0</PRIMARY_KEY>"
xml_endpoint_and_affected_ongoing_no_url += "<HOSTNAME>lhcbsrm-kit.gridka.de</HOSTNAME>"
xml_endpoint_and_affected_ongoing_no_url += "<SERVICE_TYPE>SRM</SERVICE_TYPE>"
xml_endpoint_and_affected_ongoing_no_url += "<ENDPOINT>lhcbsrm-kit.gridka.deSRM</ENDPOINT>"
xml_endpoint_and_affected_ongoing_no_url += "<HOSTED_BY>FZK-LCG2</HOSTED_BY>"
xml_endpoint_and_affected_ongoing_no_url += "<GOCDB_PORTAL_URL>https://goc.egi.eu/bof</GOCDB_PORTAL_URL>"
xml_endpoint_and_affected_ongoing_no_url += "<AFFECTED_ENDPOINTS/>"
xml_endpoint_and_affected_ongoing_no_url += "<SEVERITY>OUTAGE</SEVERITY>"
xml_endpoint_and_affected_ongoing_no_url += "<DESCRIPTION>Namespace reordering</DESCRIPTION>"
xml_endpoint_and_affected_ongoing_no_url += "<INSERT_DATE>1595233003</INSERT_DATE>"
xml_endpoint_and_affected_ongoing_no_url += "<START_DATE>1595314800</START_DATE>"
xml_endpoint_and_affected_ongoing_no_url += "<END_DATE>1595343600</END_DATE>"
xml_endpoint_and_affected_ongoing_no_url += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
xml_endpoint_and_affected_ongoing_no_url += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
xml_endpoint_and_affected_ongoing_no_url += "</DOWNTIME></ROOT>\n"


xml_endpoint_and_affected_ongoing_2_endpoints = '<?xml version="1.0"?>\n<ROOT>'
xml_endpoint_and_affected_ongoing_2_endpoints += '<DOWNTIME ID="29109" PRIMARY_KEY="123" CLASSIFICATION="SCHEDULED">'
xml_endpoint_and_affected_ongoing_2_endpoints += "<PRIMARY_KEY>123</PRIMARY_KEY>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<HOSTNAME>appsgrycap.i3m.upv.es</HOSTNAME>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<SERVICE_TYPE>es.upv.grycap.im</SERVICE_TYPE>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<ENDPOINT>appsgrycap.i3m.upv.eses.upv.grycap.im</ENDPOINT>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<HOSTED_BY>UPV-GRyCAP</HOSTED_BY>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<GOCDB_PORTAL_URL>https://goc.egi.eu/bof</GOCDB_PORTAL_URL>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<AFFECTED_ENDPOINTS>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<ENDPOINT>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<ID>6735</ID>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<NAME>IM Web portal</NAME>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<URL>https://appsgrycap.i3m.upv.es:31443/im-web/</URL>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<INTERFACENAME>es.upv.grycap.im</INTERFACENAME>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<ENDPOINT_MONITORED>N</ENDPOINT_MONITORED>"
xml_endpoint_and_affected_ongoing_2_endpoints += "</ENDPOINT>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<ENDPOINT>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<ID>6737</ID>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<NAME>IM REST API</NAME>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<URL>https://appsgrycap.i3m.upv.es:31443/im/</URL>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<INTERFACENAME>es.upv.grycap.im</INTERFACENAME>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<ENDPOINT_MONITORED>N</ENDPOINT_MONITORED>"
xml_endpoint_and_affected_ongoing_2_endpoints += "</ENDPOINT>"
xml_endpoint_and_affected_ongoing_2_endpoints += "</AFFECTED_ENDPOINTS>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<SEVERITY>OUTAGE</SEVERITY>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<DESCRIPTION>maintenance of the facilities</DESCRIPTION>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<INSERT_DATE>1594986508</INSERT_DATE>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<START_DATE>1596196800</START_DATE>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<END_DATE>1598961600</END_DATE>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
xml_endpoint_and_affected_ongoing_2_endpoints += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
xml_endpoint_and_affected_ongoing_2_endpoints += "</DOWNTIME></ROOT>\n"

XML_nodesite_ongoing = '<?xml version="1.0"?>\n<ROOT><DOWNTIME ID="78505455" PRIMARY_KEY="28490G0"'
XML_nodesite_ongoing += ' CLASSIFICATION="SCHEDULED"><HOSTNAME>egse-cresco.portici.enea.it</HOSTNAME>'
XML_nodesite_ongoing += "<HOSTED_BY>GRISU-ENEA-GRID</HOSTED_BY>"
XML_nodesite_ongoing += "<SEVERITY>OUTAGE</SEVERITY>"
XML_nodesite_ongoing += "<DESCRIPTION>Software problems</DESCRIPTION>"
XML_nodesite_ongoing += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_nodesite_ongoing += "<START_DATE>1276360500</START_DATE>"
XML_nodesite_ongoing += "<END_DATE>1276878660</END_DATE>"
XML_nodesite_ongoing += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
XML_nodesite_ongoing += "<FORMATED_END_DATE>" + nowPlus8h + "</FORMATED_END_DATE>"
XML_nodesite_ongoing += '</DOWNTIME><DOWNTIME ID="78505456" PRIMARY_KEY="28490G0"'
XML_nodesite_ongoing += ' CLASSIFICATION="SCHEDULED"><SITENAME>GRISU-ENEA-GRID</SITENAME>'
XML_nodesite_ongoing += "<SEVERITY>OUTAGE</SEVERITY>"
XML_nodesite_ongoing += "<DESCRIPTION>Software problems SITE</DESCRIPTION>"
XML_nodesite_ongoing += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_nodesite_ongoing += "<START_DATE>1276360500</START_DATE>"
XML_nodesite_ongoing += "<END_DATE>1276878660</END_DATE>"
XML_nodesite_ongoing += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
XML_nodesite_ongoing += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
XML_nodesite_ongoing += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_nodesite_ongoing += "</DOWNTIME></ROOT>\n"

XML_site_startingIn8h = '<?xml version="1.0"?>\n<ROOT><DOWNTIME ID="78505456" PRIMARY_KEY="28490G0"'
XML_site_startingIn8h += ' CLASSIFICATION="SCHEDULED"><SITENAME>GRISU-ENEA-GRID</SITENAME>'
XML_site_startingIn8h += "<SEVERITY>OUTAGE</SEVERITY>"
XML_site_startingIn8h += "<DESCRIPTION>Software problems SITE</DESCRIPTION>"
XML_site_startingIn8h += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_site_startingIn8h += "<START_DATE>1276360500</START_DATE>"
XML_site_startingIn8h += "<END_DATE>1276878660</END_DATE>"
XML_site_startingIn8h += "<FORMATED_START_DATE>" + nowPlus8h + "</FORMATED_START_DATE>"
XML_site_startingIn8h += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
XML_site_startingIn8h += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_site_startingIn8h += "</DOWNTIME></ROOT>\n"

XML_node_startingIn8h = '<?xml version="1.0"?>\n<ROOT><DOWNTIME ID="78505455" PRIMARY_KEY="28490G0"'
XML_node_startingIn8h += ' CLASSIFICATION="SCHEDULED">'
XML_node_startingIn8h += "<HOSTNAME>egse-cresco.portici.enea.it</HOSTNAME>"
XML_node_startingIn8h += "<HOSTED_BY>GRISU-ENEA-GRID</HOSTED_BY>"
XML_node_startingIn8h += "<SEVERITY>OUTAGE</SEVERITY>"
XML_node_startingIn8h += "<DESCRIPTION>Software problems</DESCRIPTION>"
XML_node_startingIn8h += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_node_startingIn8h += "<START_DATE>1276360500</START_DATE>"
XML_node_startingIn8h += "<END_DATE>1276878660</END_DATE>"
XML_node_startingIn8h += "<FORMATED_START_DATE>" + nowPlus8h + "</FORMATED_START_DATE>"
XML_node_startingIn8h += "<FORMATED_END_DATE>" + nowPlus24h + "</FORMATED_END_DATE>"
XML_node_startingIn8h += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_node_startingIn8h += "</DOWNTIME></ROOT>\n"

XML_site_ongoing_and_site_starting_in_24_hours = '<?xml version="1.0"?>\n<ROOT><DOWNTIME ID="78505456"'
XML_site_ongoing_and_site_starting_in_24_hours += ' PRIMARY_KEY="28490G1" CLASSIFICATION="SCHEDULED">'
XML_site_ongoing_and_site_starting_in_24_hours += "<SITENAME>GRISU-ENEA-GRID</SITENAME>"
XML_site_ongoing_and_site_starting_in_24_hours += "<SEVERITY>OUTAGE</SEVERITY>"
XML_site_ongoing_and_site_starting_in_24_hours += "<DESCRIPTION>Software problems SITE</DESCRIPTION>"
XML_site_ongoing_and_site_starting_in_24_hours += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<START_DATE>1276360500</START_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<END_DATE>1276878660</END_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<FORMATED_END_DATE>" + nowPlus8h + "</FORMATED_END_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_site_ongoing_and_site_starting_in_24_hours += "</DOWNTIME>"
XML_site_ongoing_and_site_starting_in_24_hours += '<DOWNTIME ID="78505457" PRIMARY_KEY="28490G0"'
XML_site_ongoing_and_site_starting_in_24_hours += ' CLASSIFICATION="SCHEDULED">'
XML_site_ongoing_and_site_starting_in_24_hours += "<SITENAME>GRISU-ENEA-GRID</SITENAME>"
XML_site_ongoing_and_site_starting_in_24_hours += "<SEVERITY>OUTAGE</SEVERITY>"
XML_site_ongoing_and_site_starting_in_24_hours += "<DESCRIPTION>Software problems SITE 2</DESCRIPTION>"
XML_site_ongoing_and_site_starting_in_24_hours += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<START_DATE>1276360500</START_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<END_DATE>1276878660</END_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<FORMATED_START_DATE>" + nowPlus24h + "</FORMATED_START_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<FORMATED_END_DATE>" + nowPlus40h + "</FORMATED_END_DATE>"
XML_site_ongoing_and_site_starting_in_24_hours += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_site_ongoing_and_site_starting_in_24_hours += "</DOWNTIME></ROOT>\n"

XML_site_startingIn24h_and_site_startingIn50h = '<?xml version="1.0"?>\n<ROOT>'
XML_site_startingIn24h_and_site_startingIn50h += '<DOWNTIME ID="78505456" PRIMARY_KEY="28490G1"'
XML_site_startingIn24h_and_site_startingIn50h += ' CLASSIFICATION="SCHEDULED">'
XML_site_startingIn24h_and_site_startingIn50h += "<SITENAME>GRISU-ENEA-GRID</SITENAME>"
XML_site_startingIn24h_and_site_startingIn50h += "<SEVERITY>OUTAGE</SEVERITY>"
XML_site_startingIn24h_and_site_startingIn50h += "<DESCRIPTION>Software problems SITE</DESCRIPTION>"
XML_site_startingIn24h_and_site_startingIn50h += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<START_DATE>1276360500</START_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<END_DATE>1276878660</END_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<FORMATED_START_DATE>" + nowPlus24h + "</FORMATED_START_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<FORMATED_END_DATE>" + nowPlus40h + "</FORMATED_END_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_site_startingIn24h_and_site_startingIn50h += '</DOWNTIME><DOWNTIME ID="78505457" PRIMARY_KEY="28490G0"'
XML_site_startingIn24h_and_site_startingIn50h += ' CLASSIFICATION="SCHEDULED">'
XML_site_startingIn24h_and_site_startingIn50h += "<SITENAME>GRISU-ENEA-GRID</SITENAME>"
XML_site_startingIn24h_and_site_startingIn50h += "<SEVERITY>OUTAGE</SEVERITY>"
XML_site_startingIn24h_and_site_startingIn50h += "<DESCRIPTION>Software problems SITE</DESCRIPTION>"
XML_site_startingIn24h_and_site_startingIn50h += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<START_DATE>1276360500</START_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<END_DATE>1276878660</END_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<FORMATED_START_DATE>" + nowPlus50h + "</FORMATED_START_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<FORMATED_END_DATE>" + nowPlus60h + "</FORMATED_END_DATE>"
XML_site_startingIn24h_and_site_startingIn50h += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_site_startingIn24h_and_site_startingIn50h += "</DOWNTIME></ROOT>\n"

XML_site_ongoing_and_other_site_starting_in_24_hours = '<?xml version="1.0"?>\n<ROOT>'
XML_site_ongoing_and_other_site_starting_in_24_hours += '<DOWNTIME ID="78505456" PRIMARY_KEY="28490G1"'
XML_site_ongoing_and_other_site_starting_in_24_hours += ' CLASSIFICATION="SCHEDULED">'
XML_site_ongoing_and_other_site_starting_in_24_hours += "<SITENAME>GRISU-ENEA-GRID</SITENAME>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<SEVERITY>OUTAGE</SEVERITY>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<DESCRIPTION>Software problems SITE</DESCRIPTION>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<START_DATE>1276360500</START_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<END_DATE>1276878660</END_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<FORMATED_END_DATE>" + nowPlus8h + "</FORMATED_END_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "</DOWNTIME>"
XML_site_ongoing_and_other_site_starting_in_24_hours += '<DOWNTIME ID="78505457" PRIMARY_KEY="28490G0"'
XML_site_ongoing_and_other_site_starting_in_24_hours += ' CLASSIFICATION="SCHEDULED">'
XML_site_ongoing_and_other_site_starting_in_24_hours += "<SITENAME>CERN-PROD</SITENAME>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<SEVERITY>OUTAGE</SEVERITY>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<DESCRIPTION>Software problems SITE 2</DESCRIPTION>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<START_DATE>1276360500</START_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<END_DATE>1276878660</END_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<FORMATED_START_DATE>" + nowPlus24h + "</FORMATED_START_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<FORMATED_END_DATE>" + nowPlus40h + "</FORMATED_END_DATE>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_site_ongoing_and_other_site_starting_in_24_hours += "</DOWNTIME></ROOT>\n"

XML_node_ongoing_and_other_node_starting_in_24_hours = '<?xml version="1.0"?>\n<ROOT>'
XML_node_ongoing_and_other_node_starting_in_24_hours += '<DOWNTIME ID="78505456" PRIMARY_KEY="28490G1"'
XML_node_ongoing_and_other_node_starting_in_24_hours += ' CLASSIFICATION="SCHEDULED">'
XML_node_ongoing_and_other_node_starting_in_24_hours += "<HOSTNAME>egse-cresco.portici.enea.it</HOSTNAME>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<HOSTED_BY>GRISU-ENEA-GRID</HOSTED_BY>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<SEVERITY>OUTAGE</SEVERITY>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<DESCRIPTION>Software problems RESOURCE</DESCRIPTION>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<START_DATE>1276360500</START_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<END_DATE>1276878660</END_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<FORMATED_START_DATE>" + nowLess12h + "</FORMATED_START_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<FORMATED_END_DATE>" + nowPlus8h + "</FORMATED_END_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_node_ongoing_and_other_node_starting_in_24_hours += '</DOWNTIME><DOWNTIME ID="78505457" PRIMARY_KEY="28490G0"'
XML_node_ongoing_and_other_node_starting_in_24_hours += ' CLASSIFICATION="SCHEDULED">'
XML_node_ongoing_and_other_node_starting_in_24_hours += "<HOSTNAME>ce112.cern.ch</HOSTNAME>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<HOSTED_BY>CERN-PROD</HOSTED_BY>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<SEVERITY>OUTAGE</SEVERITY>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<DESCRIPTION>Software problems RESOURCE 2</DESCRIPTION>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<INSERT_DATE>1276273965</INSERT_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<START_DATE>1276360500</START_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<END_DATE>1276878660</END_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<FORMATED_START_DATE>" + nowPlus24h + "</FORMATED_START_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<FORMATED_END_DATE>" + nowPlus40h + "</FORMATED_END_DATE>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "<GOCDB_PORTAL_URL>https://bof</GOCDB_PORTAL_URL>"
XML_node_ongoing_and_other_node_starting_in_24_hours += "</DOWNTIME></ROOT>\n"


# test


def test__downTimeXMLParsing_affected():

    res = GOCCli._downTimeXMLParsing(xml_endpoint_and_affected_ongoing, "Resource")
    assert set(res) == {"109962G0 lhcbsrm-kit.gridka.deSRM"}
    assert res["109962G0 lhcbsrm-kit.gridka.deSRM"]["HOSTNAME"] == "lhcbsrm-kit.gridka.de"
    assert res["109962G0 lhcbsrm-kit.gridka.deSRM"]["URL"] == "lhcbsrm-disk-kit.gridka.de"
    res = GOCCli._downTimeXMLParsing(xml_endpoint_and_affected_ongoing, "Site")
    assert res == {}

    res = GOCCli._downTimeXMLParsing(xml_endpoint_and_affected_ongoing_broken, "Resource")
    assert list(res)[0] == "109962G0 lhcbsrm-kit.gridka.deSRM"
    assert res["109962G0 lhcbsrm-kit.gridka.deSRM"]["HOSTNAME"] == "lhcbsrm-kit.gridka.de"
    res = GOCCli._downTimeXMLParsing(xml_endpoint_and_affected_ongoing_broken, "Site")
    assert res == {}

    res = GOCCli._downTimeXMLParsing(xml_endpoint_and_affected_ongoing_no_url, "Resource")
    assert list(res)[0] == "109962G0 lhcbsrm-kit.gridka.deSRM"
    assert res["109962G0 lhcbsrm-kit.gridka.deSRM"]["HOSTNAME"] == "lhcbsrm-kit.gridka.de"
    res = GOCCli._downTimeXMLParsing(xml_endpoint_and_affected_ongoing_no_url, "Site")
    assert res == {}

    res = GOCCli._downTimeXMLParsing(xml_endpoint_and_affected_ongoing_2_endpoints, "Resource")
    assert list(res)[0] == "123 appsgrycap.i3m.upv.eses.upv.grycap.im"
    assert res["123 appsgrycap.i3m.upv.eses.upv.grycap.im"]["HOSTNAME"] == "appsgrycap.i3m.upv.es"
    assert res["123 appsgrycap.i3m.upv.eses.upv.grycap.im"]["URL"] == "https://appsgrycap.i3m.upv.es:31443/im-web/"
    res = GOCCli._downTimeXMLParsing(xml_endpoint_and_affected_ongoing_2_endpoints, "Site")
    assert res == {}


def test__downTimeXMLParsing():

    res = GOCCli._downTimeXMLParsing(XML_site_ongoing, "Site")
    assert set(res) == {"28490G0 GRISU-ENEA-GRID"}
    assert res["28490G0 GRISU-ENEA-GRID"]["SITENAME"] == "GRISU-ENEA-GRID"

    res = GOCCli._downTimeXMLParsing(XML_node_ongoing, "Resource")
    assert set(res) == {"28490G0 egse-cresco.portici.enea.it"}
    assert res["28490G0 egse-cresco.portici.enea.it"]["HOSTNAME"] == "egse-cresco.portici.enea.it"
    assert res["28490G0 egse-cresco.portici.enea.it"]["HOSTED_BY"] == "GRISU-ENEA-GRID"

    res = GOCCli._downTimeXMLParsing(XML_site_ongoing, "Resource")
    assert res == {}
    res = GOCCli._downTimeXMLParsing(XML_node_ongoing, "Site")
    assert res == {}

    res = GOCCli._downTimeXMLParsing(XML_nodesite_ongoing, "Site")
    assert len(res) == 1
    assert set(res) == {"28490G0 GRISU-ENEA-GRID"}
    assert res["28490G0 GRISU-ENEA-GRID"]["SITENAME"] == "GRISU-ENEA-GRID"

    res = GOCCli._downTimeXMLParsing(XML_nodesite_ongoing, "Resource")
    assert len(res) == 1
    assert set(res) == {"28490G0 egse-cresco.portici.enea.it"}
    assert res["28490G0 egse-cresco.portici.enea.it"]["HOSTNAME"] == "egse-cresco.portici.enea.it"

    res = GOCCli._downTimeXMLParsing(XML_site_startingIn8h, "Site", None, now)
    assert res == {}
    res = GOCCli._downTimeXMLParsing(XML_node_startingIn8h, "Resource", None, now)
    assert res == {}

    res = GOCCli._downTimeXMLParsing(XML_site_ongoing_and_site_starting_in_24_hours, "Site", None, now)
    assert set(res) == {"28490G1 GRISU-ENEA-GRID"}
    assert res["28490G1 GRISU-ENEA-GRID"]["SITENAME"] == "GRISU-ENEA-GRID"

    res = GOCCli._downTimeXMLParsing(XML_site_ongoing_and_site_starting_in_24_hours, "Resource", None, now)
    assert res == {}
    res = GOCCli._downTimeXMLParsing(XML_site_startingIn24h_and_site_startingIn50h, "Site", None, now)
    assert res == {}

    res = GOCCli._downTimeXMLParsing(XML_site_startingIn24h_and_site_startingIn50h, "Site", None, tomorrow)
    assert set(res) == {"28490G1 GRISU-ENEA-GRID"}
    assert res["28490G1 GRISU-ENEA-GRID"]["SITENAME"] == "GRISU-ENEA-GRID"

    res = GOCCli._downTimeXMLParsing(XML_site_ongoing_and_other_site_starting_in_24_hours, "Site", ["GRISU-ENEA-GRID"])
    assert list(res)[0], "28490G1 GRISU-ENEA-GRID"
    assert res["28490G1 GRISU-ENEA-GRID"]["SITENAME"], "GRISU-ENEA-GRID"
    res = GOCCli._downTimeXMLParsing(
        XML_site_ongoing_and_other_site_starting_in_24_hours, "Site", ["GRISU-ENEA-GRID", "CERN-PROD"]
    )
    assert "28490G1 GRISU-ENEA-GRID" in res
    assert "28490G0 CERN-PROD" in res
    assert res["28490G1 GRISU-ENEA-GRID"]["SITENAME"] == "GRISU-ENEA-GRID"
    assert res["28490G0 CERN-PROD"]["SITENAME"] == "CERN-PROD"
    res = GOCCli._downTimeXMLParsing(XML_site_ongoing_and_other_site_starting_in_24_hours, "Site", "CERN-PROD")
    assert set(res) == {"28490G0 CERN-PROD"}
    assert res["28490G0 CERN-PROD"]["SITENAME"] == "CERN-PROD"
    res = GOCCli._downTimeXMLParsing(XML_site_ongoing_and_other_site_starting_in_24_hours, "Site", "CNAF-T1")
    assert res == {}

    res = GOCCli._downTimeXMLParsing(
        XML_site_ongoing_and_other_site_starting_in_24_hours, "Site", ["GRISU-ENEA-GRID", "CERN-PROD"], now
    )
    assert set(res) == {"28490G1 GRISU-ENEA-GRID"}
    assert res["28490G1 GRISU-ENEA-GRID"]["SITENAME"] == "GRISU-ENEA-GRID"
    res = GOCCli._downTimeXMLParsing(
        XML_site_ongoing_and_other_site_starting_in_24_hours, "Site", ["GRISU-ENEA-GRID", "CERN-PROD"], inAWeek
    )
    assert set(res) == {"28490G0 CERN-PROD", "28490G1 GRISU-ENEA-GRID"}
    assert res["28490G0 CERN-PROD"]["SITENAME"] == "CERN-PROD"

    res = GOCCli._downTimeXMLParsing(
        XML_node_ongoing_and_other_node_starting_in_24_hours, "Resource", ["egse-cresco.portici.enea.it"]
    )
    assert set(res) == {"28490G1 egse-cresco.portici.enea.it"}
    assert res["28490G1 egse-cresco.portici.enea.it"]["HOSTNAME"] == "egse-cresco.portici.enea.it"
    res = GOCCli._downTimeXMLParsing(
        XML_node_ongoing_and_other_node_starting_in_24_hours,
        "Resource",
        ["egse-cresco.portici.enea.it", "ce112.cern.ch"],
    )
    assert "28490G1 egse-cresco.portici.enea.it" in res
    assert "28490G0 ce112.cern.ch" in res
    assert res["28490G1 egse-cresco.portici.enea.it"]["HOSTNAME"] == "egse-cresco.portici.enea.it"
    assert res["28490G0 ce112.cern.ch"]["HOSTNAME"] == "ce112.cern.ch"
    res = GOCCli._downTimeXMLParsing(XML_node_ongoing_and_other_node_starting_in_24_hours, "Resource", "ce112.cern.ch")
    assert set(res) == {"28490G0 ce112.cern.ch"}
    assert res["28490G0 ce112.cern.ch"]["HOSTNAME"] == "ce112.cern.ch"
    res = GOCCli._downTimeXMLParsing(
        XML_node_ongoing_and_other_node_starting_in_24_hours, "Resource", "grid0.fe.infn.it"
    )
    assert res == {}

    res = GOCCli._downTimeXMLParsing(
        XML_node_ongoing_and_other_node_starting_in_24_hours,
        "Resource",
        ["egse-cresco.portici.enea.it", "ce112.cern.ch"],
        now,
    )
    assert "28490G1 egse-cresco.portici.enea.it" in res
    assert res["28490G1 egse-cresco.portici.enea.it"]["HOSTNAME"] == "egse-cresco.portici.enea.it"
    res = GOCCli._downTimeXMLParsing(
        XML_node_ongoing_and_other_node_starting_in_24_hours,
        "Resource",
        ["egse-cresco.portici.enea.it", "ce112.cern.ch"],
        inAWeek,
    )
    assert set(res) == {"28490G1 egse-cresco.portici.enea.it", "28490G0 ce112.cern.ch"}
    assert res["28490G0 ce112.cern.ch"]["HOSTNAME"] == "ce112.cern.ch"
